"""The ``input`` module allows you to access the conversion routines programmatically

To use the routines, you need to instantiate a ``ReadStrategy`` and a ``WriteStrategy``
relevant for the format of the input and output data.  You then pass these to a
``Context``.

Example
-------
Convert an in-memory dictionary of pandas DataFrames containing OSeMOSYS parameters
to an Excel spreadsheet::

>>> from otoole import ReadMemory
>>> from otoole import WriteExcel
>>> from otoole import Context
>>> reader = ReadMemory(parameters)
>>> writer = WriteExcel()
>>> converter = Context(read_strategy=reader, write_strategy=writer)
>>> converter.convert('.', 'osemosys_to_excel.xlsx')

Convert a GNUMathProg datafile to a folder of CSV files::

>>> from otoole import ReadDataFile
>>> from otoole import WriteCsv
>>> from otoole import Context
>>> reader = ReadDataFile()
>>> writer = WriteCsv()
>>> converter = Context(read_strategy=reader, write_strategy=writer)
>>> converter.convert('my_datafile.txt', 'folder_of_csv_files')

"""
from __future__ import annotations

import logging
from abc import ABC, abstractmethod
from typing import Any, Dict, Optional, TextIO, Tuple, Union

import pandas as pd

logger = logging.getLogger(__name__)


class Context:
    """
    The Context defines the interface of interest to clients.
    """

    def __init__(
        self, read_strategy: ReadStrategy, write_strategy: WriteStrategy
    ) -> None:
        """
        Usually, the Context accepts a strategy through the constructor, but
        also provides a setter to change it at runtime.
        """
        self._read_strategy = read_strategy
        self._write_strategy = write_strategy

    @property
    def write_strategy(self) -> WriteStrategy:
        """
        The Context maintains a reference to one of the Strategy objects. The
        Context does not know the concrete class of a strategy. It should work
        with all strategies via the Strategy interface.
        """
        return self._write_strategy

    @write_strategy.setter
    def write_strategy(self, strategy: WriteStrategy) -> None:
        """
        Usually, the Context allows replacing a Strategy object at runtime.
        """
        self._write_strategy = strategy

    @property
    def read_strategy(self) -> ReadStrategy:
        """
        The Context maintains a reference to one of the Strategy objects. The
        Context does not know the concrete class of a strategy. It should work
        with all strategies via the Strategy interface.
        """
        return self._read_strategy

    @read_strategy.setter
    def read_strategy(self, strategy: ReadStrategy) -> None:
        """
        Usually, the Context allows replacing a Strategy object at runtime.
        """
        self._read_strategy = strategy

    def _read(
        self, filepath: str, **kwargs
    ) -> Tuple[Dict[str, pd.DataFrame], Dict[str, Any]]:
        """Delegate reading to the strategy, depending upon the format"""
        return self._read_strategy.read(filepath, **kwargs)

    def _write(
        self, inputs: Dict, filepath: str, default_values: Dict, **kwargs
    ) -> None:
        """
        Delegate writing to the strategy, depending upon the format
        """
        self._write_strategy.write(inputs, filepath, default_values, **kwargs)

    def convert(self, input_filepath: str, output_filepath: str, **kwargs: Dict):
        """Converts from file ``input_filepath`` to file ``output_filepath``

        Arguments
        ---------
        input_filepath: str
        output_filepath: str
        """
        inputs, default_values = self._read(input_filepath, **kwargs)
        self._write(inputs, output_filepath, default_values, **kwargs)


class Strategy(ABC):
    """

    Arguments
    ---------
    user_config : dict, default=None
        A user configuration for the input parameters and sets
    """

    def __init__(self, user_config: Dict[str, Dict]):

        self.user_config = user_config

        # self.input_config = {
        #     x: y for x, y in self.user_config.items() if y["type"] in ["param", 'set']
        # }
        self.results_config = {
            x: y for x, y in self.user_config.items() if y["type"] == "result"
        }

    def _add_dtypes(self, config: Dict):
        for name, details in config.items():
            if details["type"] == "param":
                dtypes = {}
                for column in details["indices"] + ["VALUE"]:
                    if column == "VALUE":
                        dtypes["VALUE"] = details["dtype"]
                    else:
                        dtypes[column] = config[column]["dtype"]
                details["index_dtypes"] = dtypes
        return config

    @property
    def user_config(self) -> Dict:
        return self._user_config

    @user_config.setter
    def user_config(self, value: Dict):
        if value:
            self._user_config = self._add_dtypes(value)
        elif value is None:
            raise ValueError("A user configuration must be passed into the reader")

    @staticmethod
    def _read_default_values(config):
        default_values = {}
        for name, contents in config.items():
            if contents["type"] != "set":
                default_values[name] = contents["default"]
        return default_values


class WriteStrategy(Strategy):
    """
    The WriteStrategy interface declares operations common to all writing formats

    The Context uses this interface to call the algorithm defined by Concrete
    Strategies.

    Arguments
    ---------
    filepath: str, default=None
    default_values: dict, default=None
    user_config: dict, default=None

    """

    def __init__(
        self,
        user_config: Dict,
        filepath: Optional[str] = None,
        default_values: Optional[Dict] = None,
        write_defaults: bool = False,
        input_data: Optional[Dict[str, pd.DataFrame]] = None,
    ):
        super().__init__(user_config=user_config)
        if filepath:
            self.filepath = filepath
        else:
            self.filepath = ""

        if default_values:
            self.default_values = default_values
        else:
            self.default_values = {}

        if input_data:
            self.input_data = input_data
        else:
            self.input_data = {}

        self.write_defaults = write_defaults

    @abstractmethod
    def _header(self) -> Union[TextIO, Any]:
        raise NotImplementedError()

    @abstractmethod
    def _write_parameter(
        self,
        df: pd.DataFrame,
        parameter_name: str,
        handle: TextIO,
        default: float,
        **kwargs,
    ) -> pd.DataFrame:
        """Write parameter data"""
        raise NotImplementedError()

    @abstractmethod
    def _write_set(self, df: pd.DataFrame, set_name, handle: TextIO) -> pd.DataFrame:
        """Write set data"""
        raise NotImplementedError()

    @abstractmethod
    def _footer(self, handle: TextIO):
        raise NotImplementedError()

    def write(
        self,
        inputs: Dict[str, pd.DataFrame],
        filepath: str,
        default_values: Dict[str, float],
        **kwargs,
    ):
        """Perform the conversion from dict of dataframes to destination format"""
        self.filepath = filepath
        self.default_values = default_values

        handle = self._header()
        logger.debug(default_values)

        if self.write_defaults:
            try:
                inputs = self._expand_defaults(inputs, default_values, **kwargs)
            except KeyError as ex:
                logger.debug(ex)

        for name, df in sorted(inputs.items()):
            logger.debug("%s has %s columns: %s", name, len(df.index.names), df.columns)

            try:
                entity_type = self.user_config[name]["type"]
            except KeyError:
                try:
                    entity_type = self.results_config[name]["type"]
                except KeyError:
                    raise KeyError("Cannot find %s in input or results config", name)

            if entity_type != "set":
                default_value = default_values[name]
                self._write_parameter(
                    df, name, handle, default=default_value, input_data=self.input_data
                )
            else:
                self._write_set(df, name, handle)

        self._footer(handle)

        if isinstance(handle, TextIO):
            handle.close()

    def _expand_defaults(
        self,
        data_to_expand: Dict[str, pd.DataFrame],
        default_values: Dict[str, float],
        **kwargs,
    ) -> Dict[str, pd.DataFrame]:
        """Populates default value entry rows in dataframes

        Parameters
        ----------
        input_data : Dict[str, pd.DataFrame],
        default_values : Dict[str, float]

        Returns
        -------
        results : Dict[str, pd.DataFrame]
            Updated available reults dictionary

        Raises
        ------
        KeyError
            If set defenitons are not in input_data and input_data is not supplied
        """

        sets = [x for x in self.user_config if self.user_config[x]["type"] == "set"]

        # if expanding results, input data is needed for set defenitions
        if "input_data" in kwargs:
            model_data = kwargs["input_data"]
        else:
            model_data = data_to_expand

        output = {}
        for name, data in data_to_expand.items():
            logger.info(f"Writing defaults for {name}")

            # skip sets
            if name in sets:
                output[name] = data
                continue

            # TODO
            # Issue with how otoole handles trade route right now.
            # The double defenition of REGION throws an error.
            if name == "TradeRoute":
                output[name] = data
                continue

            # save set information for each parameter
            index_data = {}
            for index in data.index.names:
                try:
                    index_data[index] = model_data[index]["VALUE"].to_list()
                except KeyError as ex:
                    logger.info("Can not write default values. Supply input data")
                    raise KeyError(ex)

            # set index
            if len(index_data) > 1:
                new_index = pd.MultiIndex.from_product(
                    list(index_data.values()), names=list(index_data.keys())
                )
            else:
                new_index = pd.Index(
                    list(index_data.values())[0], name=list(index_data.keys())[0]
                )
            df_default = pd.DataFrame(index=new_index)

            # save default result value
            df_default["VALUE"] = default_values[name]

            # combine result and default value dataframe
            df = pd.concat([data, df_default])
            df = df[~df.index.duplicated(keep="first")]
            df = df.sort_index()
            output[name] = df

        return output


class ReadStrategy(Strategy):
    """
    The Strategy interface declares operations common to all reading formats.

    The Context uses this interface to call the algorithm defined by Concrete
    Strategies.
    """

    def _check_index(
        self, input_data: Dict[str, pd.DataFrame]
    ) -> Dict[str, pd.DataFrame]:
        """Checks index and datatypes are applied to the parameter DataFrame

        Also removes empty lines

        Arguments
        ---------
        input_data : dict
            Dictionary and pandas DataFrames containing the OSeMOSYS parameters

        Returns
        -------
        dict
            Dictionary and pandas DataFrames containing the OSeMOSYS parameters
        """
        for name, df in input_data.items():

            details = self.user_config[name]

            if details["type"] == "param":
                logger.debug("Identified {} as a parameter".format(name))
                try:
                    df.set_index(details["indices"], inplace=True)
                except KeyError:
                    logger.debug("Unable to set index on {}".format(name))
                    pass

                logger.debug(
                    "Column dtypes identified: {}".format(details["index_dtypes"])
                )
                logger.debug(df.head())
                # Drop empty rows
                df = (
                    df.dropna(axis=0, how="all")
                    .reset_index()
                    .astype(details["index_dtypes"])
                    .set_index(details["indices"])
                )
            else:
                logger.debug("Identified {} as a set".format(name))
                df = df.astype(details["dtype"])

            input_data[name] = df

        return input_data

    def _get_missing_input_dataframes(
        self, input_data: Dict[str, pd.DataFrame], config_type: str
    ) -> Dict[str, pd.DataFrame]:
        """Creates empty dataframes if user config data does not exist

        Arguments:
        ----------
        input_data: Dict[str, pd.DataFrame]
            Data read in from the excel notebook
        config_type: str
            Type of value. Must be "set", "param", or "result"

        Returns:
        --------
        all_params: Dict[str, pd.DataFrame]
            Input data plus empty dataframes
        """

        if config_type not in ["set", "param", "result"]:
            raise ValueError(f"{config_type} not of type 'set', 'param', or 'result'")

        all_values = [
            value
            for value, data in self.user_config.items()
            if data["type"] == config_type
        ]
        missing_values = [x for x in all_values if x not in input_data]

        for value in missing_values:
            try:  # param and result condition
                indices = self.user_config[value]["indices"]
                df = pd.DataFrame(columns=indices)
                df = df.set_index(indices)
            except KeyError:  # set condition
                df = pd.DataFrame()
            df["VALUE"] = ""
            input_data[value] = df

        return input_data

    @abstractmethod
    def read(
        self, filepath: Union[str, TextIO], **kwargs
    ) -> Tuple[Dict[str, pd.DataFrame], Dict[str, Any]]:
        raise NotImplementedError()



import logging
import os
from typing import Any, Dict, List, Optional, TextIO, Tuple, Union

import pandas as pd
from amply import Amply
from flatten_dict import flatten

from otoole.exceptions import OtooleDeprecationError, OtooleExcelNameMismatchError
from otoole.input import ReadStrategy
from otoole.preprocess.longify_data import check_datatypes, check_set_datatype
from otoole.utils import create_name_mappings

logger = logging.getLogger(__name__)


class ReadMemory(ReadStrategy):
    """Read a dict of OSeMOSYS parameters from memory"""

    def __init__(
        self, parameters: Dict[str, pd.DataFrame], user_config: Dict[str, Dict]
    ):
        super().__init__(user_config)
        self._parameters = parameters

    def read(
        self, filepath: Union[str, TextIO] = None, **kwargs
    ) -> Tuple[Dict[str, pd.DataFrame], Dict[str, Any]]:

        config = self.user_config
        default_values = self._read_default_values(config)
        self._parameters = self._check_index(self._parameters)
        return self._parameters, default_values


class _ReadTabular(ReadStrategy):
    def __init__(self, user_config: Dict[str, Dict], keep_whitespace: bool = False):
        super().__init__(user_config)
        self.keep_whitespace = keep_whitespace

    def _check_set(self, df: pd.DataFrame, config_details: Dict, name: str):

        logger.info("Checking set %s", name)
        narrow = df

        return narrow

    def _check_parameter(self, df: pd.DataFrame, expected_headers: List, name: str):
        """Converts a dataframe from wide to narrow format

        Arguments
        ---------
        df: pd.DataFrame
        expected_headers: List
        name: str
        """
        actual_headers = df.columns
        logger.debug("Expected headers for %s: %s", name, expected_headers)

        if "REGION" in expected_headers and "REGION" not in actual_headers:
            raise ValueError("No REGION column provided for %s", name)

        if "MODEOFOPERATION" in actual_headers:
            df = df.rename(columns={"MODEOFOPERATION": "MODE_OF_OPERATION"})

        if actual_headers[-1] == "VALUE":
            logger.info(
                "%s is already in narrow form with headers %s", name, df.columns
            )
            narrow = df
        else:
            try:
                narrow = pd.melt(
                    df,
                    id_vars=expected_headers[:-1],
                    var_name=expected_headers[-1],  # Normally 'YEAR'
                    value_name="new_VALUE",
                )
                narrow = narrow.rename(columns={"new_VALUE": "VALUE"})
            except IndexError as ex:
                logger.debug("Could not reshape %s", df.columns)
                raise ex

        all_headers = expected_headers + ["VALUE"]
        for column in all_headers:
            if column not in narrow.columns:
                logger.warning("%s not in header of %s", column, name)

        logger.debug("Final all headers for %s: %s", name, all_headers)

        return narrow[all_headers].set_index(expected_headers)

    def _whitespace_converter(self, indices: List[str]) -> Dict[str, Any]:
        """Creates converter for striping whitespace in dataframe

        Arguments
        ---------
        indicies: List[str]
            Column headers of dataframe

        Returns
        -------
        Dict[str,Any]
            Converter dictionary
        """
        if self.keep_whitespace:
            return {}
        else:
            return {x: str.strip for x in indices}


class ReadExcel(_ReadTabular):
    """Read in an Excel spreadsheet in wide format to a dict of Pandas DataFrames"""

    def read(
        self, filepath: Union[str, TextIO], **kwargs
    ) -> Tuple[Dict[str, pd.DataFrame], Dict[str, Any]]:

        config = self.user_config
        default_values = self._read_default_values(config)
        excel_to_csv = create_name_mappings(config, map_full_to_short=False)

        xl = pd.ExcelFile(filepath, engine="openpyxl")

        self._check_input_sheet_names(xl.sheet_names)

        input_data = {}

        for name in xl.sheet_names:

            try:
                mod_name = excel_to_csv[name]
            except KeyError:
                mod_name = name

            config_details = config[mod_name]

            df = xl.parse(name)

            entity_type = config[mod_name]["type"]

            if entity_type == "param":
                narrow = self._check_parameter(df, config_details["indices"], mod_name)
            elif entity_type == "set":
                narrow = self._check_set(df, config_details, mod_name)

            input_data[mod_name] = narrow

        for config_type in ["param", "set"]:
            input_data = self._get_missing_input_dataframes(
                input_data, config_type=config_type
            )

        input_data = self._check_index(input_data)

        return input_data, default_values

    def _check_input_sheet_names(self, sheet_names: List[str]) -> None:
        """Checks that excel sheet names are in the config file.

        Arguments:
        ---------
        sheet_names: list[str]
            Sheet names from the excel file

        Raises:
        -------
        OtooleExcelNameMismatchError
            If the sheet name is not found in the config files parameter or
            'short_name' parameter
        """
        user_config = self.user_config
        csv_to_excel = create_name_mappings(user_config)
        config_param_names = []
        for name in user_config:
            try:
                config_param_names.append(csv_to_excel[name])
            except KeyError:
                config_param_names.append(name)

        for sheet_name in sheet_names:
            if sheet_name not in config_param_names:
                raise OtooleExcelNameMismatchError(excel_name=sheet_name)


class ReadCsv(_ReadTabular):
    """Read in a folder of CSV files"""

    def read(
        self, filepath, **kwargs
    ) -> Tuple[Dict[str, pd.DataFrame], Dict[str, Any]]:

        input_data = {}

        self._check_for_default_values_csv(filepath)
        default_values = self._read_default_values(self.user_config)

        for parameter, details in self.user_config.items():
            logger.info("Looking for %s", parameter)

            entity_type = details["type"]
            try:
                converter = self._whitespace_converter(details["indices"])
            except KeyError:  # sets don't have indices def
                converter = self._whitespace_converter(["VALUE"])

            if entity_type == "param":
                df = self._get_input_data(filepath, parameter, details, converter)
                narrow = self._check_parameter(df, details["indices"], parameter)
                if not narrow.empty:
                    narrow_checked = check_datatypes(
                        narrow, self.user_config, parameter
                    )
                else:
                    narrow_checked = narrow

            elif entity_type == "set":
                df = self._get_input_data(filepath, parameter, details, converter)
                narrow = self._check_set(df, details, parameter)
                if not narrow.empty:
                    narrow_checked = check_set_datatype(
                        narrow, self.user_config, parameter
                    )
                else:
                    narrow_checked = narrow

            else:  # results
                continue

            input_data[parameter] = narrow_checked

        for config_type in ["param", "set"]:
            input_data = self._get_missing_input_dataframes(
                input_data, config_type=config_type
            )

        input_data = self._check_index(input_data)

        return input_data, default_values

    @staticmethod
    def _get_input_data(
        filepath: str, parameter: str, details: Dict, converter: Optional[Dict] = None
    ) -> pd.DataFrame:
        """Reads in and checks CSV data format.

        Arguments
        ---------
        filepath:str
            Directory of csv files
        parameter:str
            parameter name
        config_details: dict[str,Union[str,float,int]]
            configuration data for the parameter being read in

        Returns
        -------
        pd.DataFrame
            CSV data as a dataframe
        """
        converter = {} if not converter else converter
        csv_path = os.path.join(filepath, parameter + ".csv")
        try:
            df = pd.read_csv(csv_path, converters=converter)
        except pd.errors.EmptyDataError:
            logger.error("No data found in file for %s", parameter)
            expected_columns = details["indices"]
            default_columns = expected_columns + ["VALUE"]
            df = pd.DataFrame(columns=default_columns)
        return df

    @staticmethod
    def _check_for_default_values_csv(filepath: str) -> None:
        """Checks for a default values csv, which has been deprecated.

        Arguments
        ---------
        filepath:str
            Directory of csv files

        Raises
        ------
        OtooleDeprecationError
            If a default_values.csv is found in input data
        """

        default_values_csv_path = os.path.join(
            os.path.dirname(filepath), "default_values.csv"
        )
        if os.path.exists(default_values_csv_path):
            raise OtooleDeprecationError(
                resource="data/default_values.csv",
                message="Remove default_values.csv and define all default values in the configuration file",
            )


class ReadDatafile(ReadStrategy):
    def read(
        self, filepath, **kwargs
    ) -> Tuple[Dict[str, pd.DataFrame], Dict[str, Any]]:

        config = self.user_config
        default_values = self._read_default_values(config)
        amply_datafile = self.read_in_datafile(filepath, config)
        inputs = self._convert_amply_to_dataframe(amply_datafile, config)
        for config_type in ["param", "set"]:
            inputs = self._get_missing_input_dataframes(inputs, config_type=config_type)
        inputs = self._check_index(inputs)
        return inputs, default_values

    def read_in_datafile(self, path_to_datafile: str, config: Dict) -> Amply:
        """Read in a datafile using the Amply parsing class

        Arguments
        ---------
        path_to_datafile: str
        config: Dict
        """
        parameter_definitions = self._load_parameter_definitions(config)
        datafile_parser = Amply(parameter_definitions)

        with open(path_to_datafile, "r") as datafile:
            datafile_parser.load_file(datafile)

        return datafile_parser

    def _load_parameter_definitions(self, config: dict) -> str:
        """Load the set and parameter dimensions into datafile parser

        Returns
        -------
        str
        """
        elements = ""

        for name, attributes in config.items():
            if attributes["type"] == "param":
                elements += "param {} {};\n".format(
                    name, "{" + ",".join(attributes["indices"]) + "}"
                )
            elif attributes["type"] == "symbolic":
                elements += "param {0} symbolic := '{1}' ;\n".format(
                    name, attributes["default"]
                )
            elif attributes["type"] == "set":
                elements += "set {};\n".format(name)

        return elements

    def _convert_amply_to_dataframe(
        self, datafile_parser: Amply, config: Dict
    ) -> Dict[str, pd.DataFrame]:
        """Converts an amply parser to a dict of pandas dataframes

        Arguments
        ---------
        datafile_parser : Amply
        config : Dict

        Returns
        -------
        dict of pandas.DataFrame
        """

        dict_of_dataframes = {}  # type: Dict[str, pd.DataFrame]

        for name in datafile_parser.symbols.keys():
            logger.debug("Extracting data for %s", name)
            if name in config and config[name]["type"] == "param":
                dict_of_dataframes[name] = self.extract_param(
                    config, name, datafile_parser, dict_of_dataframes
                )
            elif name in config and config[name]["type"] == "set":
                dict_of_dataframes[name] = self.extract_set(
                    datafile_parser, name, config, dict_of_dataframes
                )
            else:
                logger.warning(
                    "Parameter {} could not be found in the configuration.".format(name)
                )

        return dict_of_dataframes

    def extract_set(
        self, datafile_parser, name, config, dict_of_dataframes
    ) -> pd.DataFrame:
        data = datafile_parser[name].data

        indices = ["VALUE"]
        df = pd.DataFrame(data=data, columns=indices, dtype=config[name]["dtype"])

        return check_set_datatype(df, config, name)

    def extract_param(
        self, config, name, datafile_parser, dict_of_dataframes
    ) -> pd.DataFrame:
        indices = config[name]["indices"].copy()
        indices_dtypes = [config[index]["dtype"] for index in indices]
        indices.append("VALUE")
        indices_dtypes.append("float")

        raw_data = datafile_parser[name].data
        data = self._convert_amply_data_to_list(raw_data)
        df = pd.DataFrame(data=data, columns=indices)
        try:
            return check_datatypes(df, config, name)
        except ValueError as ex:
            msg = "Validation error when checking datatype of {}: {}".format(
                name, str(ex)
            )
            raise ValueError(msg)

    def _convert_amply_data_to_list(self, amply_data: Dict) -> List[List]:
        """Flattens a dictionary into a list of lists

        Arguments
        ---------
        amply_data: dict
        """

        data = []

        raw_data = flatten(amply_data)
        for key, value in raw_data.items():
            data.append(list(key) + [value])

        return data



"""Module to create template data"""
from typing import Any, Dict, Tuple

import pandas as pd

from otoole.read_strategies import ReadCsv
from otoole.utils import read_packaged_file


def get_csv_setup_data(
    config: Dict[str, Any]
) -> Tuple[Dict[str, pd.DataFrame], Dict[str, Any]]:
    """Gets template dataframe data to write as csvs

    Arguments
    ---------
    config: Dict[str,Any]
        Configuration data to get CSV data to match against

    Returns
    -------
    Dict[str: pd.DataFrame]
        Parameters with empty template dataframes
    Dict[str: Any]
        Default values extracted from config file
    """

    input_data: Dict[str, pd.DataFrame] = {}

    reader = ReadCsv(user_config=config)

    for config_type in ["param", "set"]:
        input_data = reader._get_missing_input_dataframes(
            input_data, config_type=config_type
        )
    input_data = reader._check_index(input_data)
    default_values = reader._read_default_values(config)

    return input_data, default_values


def get_config_setup_data() -> Dict[str, Any]:
    """Reads in template config data"""
    return read_packaged_file("config.yaml", "otoole.preprocess")



# -*- coding: utf-8 -*-
"""
    Dummy conftest.py for otoole.

    If you don't know what this is for, just leave it empty.
    Read more about conftest.py under:
    https://pytest.org/latest/plugins.html
"""

import os
from typing import Dict

import pandas as pd
from pytest import fixture

from otoole.utils import _read_file


@fixture
def user_config() -> Dict:
    """Reads in an example user config

    Read in an example user config file which can be passed into all
    Read and Write strategies for testing

    Returns
    -------
    Dict
    """
    file_path = os.path.join("tests", "fixtures", "config.yaml")
    with open(file_path, "r") as config_file:
        config = _read_file(config_file, ".yaml")  # typing: Dict

    file_path = os.path.join("tests", "fixtures", "config_r.yaml")
    with open(file_path, "r") as config_file:
        results = _read_file(config_file, ".yaml")  # typing: Dict

    config.update(results)

    return config


@fixture
def annual_technology_emissions_by_mode():
    df = pd.DataFrame(
        data=[["SIMPLICITY", "GAS_EXTRACTION", "CO2", 1, 2014, 1.0]],
        columns=[
            "REGION",
            "TECHNOLOGY",
            "EMISSION",
            "MODE_OF_OPERATION",
            "YEAR",
            "VALUE",
        ],
    ).set_index(["REGION", "TECHNOLOGY", "EMISSION", "MODE_OF_OPERATION", "YEAR"])
    return df


@fixture
def discount_rate():
    df = pd.DataFrame(
        data=[["SIMPLICITY", 0.05]],
        columns=["REGION", "VALUE"],
    ).set_index(["REGION"])
    return df


@fixture
def discount_rate_idv():
    df = pd.DataFrame(
        data=[["SIMPLICITY", "GAS_EXTRACTION", 0.05], ["SIMPLICITY", "DUMMY", 0.05]],
        columns=["REGION", "TECHNOLOGY", "VALUE"],
    ).set_index(["REGION", "TECHNOLOGY"])
    return df


@fixture
def discount_rate_storage():
    df = pd.DataFrame(
        data=[["SIMPLICITY", "DAM", 0.05]],
        columns=["REGION", "STORAGE", "VALUE"],
    ).set_index(["REGION", "STORAGE"])
    return df


@fixture
def emission_activity_ratio():
    df = pd.DataFrame(
        data=[["SIMPLICITY", "GAS_EXTRACTION", "CO2", 1, 2014, 1.0]],
        columns=[
            "REGION",
            "TECHNOLOGY",
            "EMISSION",
            "MODE_OF_OPERATION",
            "YEAR",
            "VALUE",
        ],
    ).set_index(["REGION", "TECHNOLOGY", "EMISSION", "MODE_OF_OPERATION", "YEAR"])
    return df


@fixture
def emission_activity_ratio_two_techs():
    df = pd.DataFrame(
        data=[
            ["SIMPLICITY", "GAS_EXTRACTION", "CO2", 1, 2014, 1.0],
            ["SIMPLICITY", "DUMMY", "CO2", 1, 2014, 0.0],
        ],
        columns=[
            "REGION",
            "TECHNOLOGY",
            "EMISSION",
            "MODE_OF_OPERATION",
            "YEAR",
            "VALUE",
        ],
    ).set_index(["REGION", "TECHNOLOGY", "EMISSION", "MODE_OF_OPERATION", "YEAR"])

    return df


@fixture
def emissions_penalty():
    df = pd.DataFrame(
        data=[
            ["SIMPLICITY", "CO2", 2014, 1.23],
            ["SIMPLICITY", "CO2", 2015, 1.23],
            ["SIMPLICITY", "CO2", 2016, 1.23],
        ],
        columns=["REGION", "EMISSION", "YEAR", "VALUE"],
    ).set_index(["REGION", "EMISSION", "YEAR"])
    return df


@fixture
def yearsplit():
    df = pd.DataFrame(
        data=[
            ["ID", 2014, 0.1667],
            ["IN", 2014, 0.0833],
            ["SD", 2014, 0.1667],
            ["SN", 2014, 0.0833],
            ["WD", 2014, 0.3333],
            ["WN", 2014, 0.1667],
        ],
        columns=["TIMESLICE", "YEAR", "VALUE"],
    ).set_index(["TIMESLICE", "YEAR"])
    return df



"""Provides a command line interface to the ``otoole`` package

The key functions are **convert**, **cplex** and **viz**.

The ``convert`` command allows convertion of multiple different OSeMOSYS input formats
including from/to csv, an AMPL format datafile, a folder of
CSVs, an Excel workbook with one tab per parameter, an SQLite database

The ``cplex`` command provides access to scripts which transform and process a CPLEX
solution file into a format which is more readily processed - either to CBC or CSV
format.

The ``validate`` command checks the technology and fuel names
against a standard or user defined configuration file.

The **viz** command allows you to produce a Reference Energy System diagram

Example
-------

Ask for help on the command line::

    >>> $ otoole --help
    usage: otoole [-h] [--verbose] [--version] {convert,cplex,validate,viz} ...

    otoole: Python toolkit of OSeMOSYS users

    positional arguments:
    {convert,cplex,validate,viz}
        convert             Convert from one input format to another
        cplex               Process a CPLEX solution file
        validate            Validate an OSeMOSYS model
        viz                 Visualise the model

    optional arguments:
    -h, --help            show this help message and exit
    --verbose, -v         Enable debug mode
    --version, -V         The version of otoole

"""
import argparse
import logging
import os
import shutil
import sys

from otoole import (
    ReadCbc,
    ReadCplex,
    ReadCsv,
    ReadDatafile,
    ReadExcel,
    ReadGurobi,
    WriteCsv,
    WriteDatafile,
    WriteExcel,
    __version__,
)
from otoole.exceptions import OtooleSetupError
from otoole.input import Context
from otoole.preprocess.setup import get_config_setup_data, get_csv_setup_data
from otoole.utils import (
    _read_file,
    read_deprecated_datapackage,
    read_packaged_file,
    validate_config,
)
from otoole.validate import main as validate
from otoole.visualise import create_res

logger = logging.getLogger(__name__)


def validate_model(args):
    data_format = args.data_format
    data_file = args.data_file
    user_config = args.user_config

    _, ending = os.path.splitext(user_config)
    with open(user_config, "r") as user_config_file:
        config = _read_file(user_config_file, ending)
    validate_config(config)

    if data_format == "datafile":
        read_strategy = ReadDatafile(user_config=config)
    elif data_format == "datapackage":
        logger.warning(
            "Reading from datapackage is deprecated, trying to read from CSVs"
        )
        data_file = read_deprecated_datapackage(data_file)
        logger.info("Successfully read folder of CSVs")
        read_strategy = ReadCsv(user_config=config)
    elif data_format == "csv":
        read_strategy = ReadCsv(user_config=config)
    elif data_format == "excel":
        read_strategy = ReadExcel(user_config=config)

    input_data, _ = read_strategy.read(data_file)

    if args.validate_config:
        validation_config = read_packaged_file(args.validate_config)
        validate(input_data, validation_config)
    else:
        validate(input_data)


def cplex2cbc(args):
    ReadCplex()._convert_cplex_file(
        args.cplex_file,
        args.output_file,
        args.start_year,
        args.end_year,
        args.output_format,
    )


def result_matrix(args):
    """Post-process results from CBC solution file into CSV format"""
    msg = "Conversion from {} to {} is not yet implemented".format(
        args.from_format, args.to_format
    )

    read_strategy = None
    write_strategy = None

    config = None
    if args.config:
        _, ending = os.path.splitext(args.config)
        with open(args.config, "r") as config_file:
            config = _read_file(config_file, ending)
        logger.info("Reading config from {}".format(args.config))
        logger.info("Validating config from {}".format(args.config))
        validate_config(config)

    # set read strategy

    if args.from_format == "cbc":
        read_strategy = ReadCbc(user_config=config)
    elif args.from_format == "cplex":
        read_strategy = ReadCplex(user_config=config)
    elif args.from_format == "gurobi":
        read_strategy = ReadGurobi(user_config=config)

    # set write strategy

    write_defaults = True if args.write_defaults else False

    if args.to_format == "csv":
        write_strategy = WriteCsv(user_config=config, write_defaults=write_defaults)

    if args.input_datapackage:
        logger.warning(
            "Reading from datapackage is deprecated, trying to read from CSVs"
        )
        input_csvs = read_deprecated_datapackage(args.input_datapackage)
        logger.info("Successfully read folder of CSVs")
        input_data, _ = ReadCsv(user_config=config).read(input_csvs)
    elif args.input_datafile:
        input_data, _ = ReadDatafile(user_config=config).read(args.input_datafile)
    else:
        input_data = {}

    if read_strategy and write_strategy:
        context = Context(read_strategy, write_strategy)
        context.convert(args.from_path, args.to_path, input_data=input_data)
    else:
        raise NotImplementedError(msg)


def conversion_matrix(args):
    """Convert from one format to another

    Implemented conversion functions::

        from\to     ex cs df
        --------------------
        excel       -- yy --
        csv         nn -- nn
        datafile    nn -- --

    """

    msg = "Conversion from {} to {} is not yet implemented".format(
        args.from_format, args.to_format
    )

    read_strategy = None
    write_strategy = None

    from_path = args.from_path
    to_path = args.to_path

    config = None
    if args.config:
        _, ending = os.path.splitext(args.config)
        with open(args.config, "r") as config_file:
            config = _read_file(config_file, ending)
        logger.info("Reading config from {}".format(args.config))
        logger.info("Validating config from {}".format(args.config))
        validate_config(config)

    # set read strategy

    keep_whitespace = True if args.keep_whitespace else False

    if args.from_format == "datafile":
        read_strategy = ReadDatafile(user_config=config)
    elif args.from_format == "datapackage":
        logger.warning(
            "Reading from datapackage is deprecated, trying to read from CSVs"
        )
        from_path = read_deprecated_datapackage(from_path)
        logger.info("Successfully read folder of CSVs")
        read_strategy = ReadCsv(user_config=config, keep_whitespace=keep_whitespace)
    elif args.from_format == "csv":
        read_strategy = ReadCsv(user_config=config, keep_whitespace=keep_whitespace)
    elif args.from_format == "excel":
        read_strategy = ReadExcel(user_config=config, keep_whitespace=keep_whitespace)

    input_data, _ = read_strategy.read(args.from_path)

    # set write strategy

    write_defaults = True if args.write_defaults else False

    if args.to_format == "datapackage":
        logger.warning("Writing to datapackage is deprecated, writing to CSVs")
        to_path = os.path.join(os.path.dirname(to_path), "data")
        write_strategy = WriteCsv(user_config=config, write_defaults=write_defaults)
    elif args.to_format == "excel":
        write_strategy = WriteExcel(
            user_config=config, write_defaults=write_defaults, input_data=input_data
        )
    elif args.to_format == "datafile":
        write_strategy = WriteDatafile(
            user_config=config, write_defaults=write_defaults
        )
    elif args.to_format == "csv":
        write_strategy = WriteCsv(user_config=config, write_defaults=write_defaults)

    if read_strategy and write_strategy:
        context = Context(read_strategy, write_strategy)
        context.convert(from_path, to_path)
    else:
        raise NotImplementedError(msg)


def data2res(args):
    """Get input data and call res creation."""

    data_format = args.data_format
    data_path = args.data_path

    _, ending = os.path.splitext(args.config)
    with open(args.config, "r") as config_file:
        config = _read_file(config_file, ending)
    validate_config(config)

    if data_format == "datafile":
        read_strategy = ReadDatafile(user_config=config)
    elif data_format == "datapackage":
        logger.warning(
            "Reading from datapackage is deprecated, trying to read from CSVs"
        )
        data_path = read_deprecated_datapackage(data_path)
        read_strategy = ReadCsv(user_config=config)
    elif data_format == "csv":
        read_strategy = ReadCsv(user_config=config)
    elif data_format == "excel":
        read_strategy = ReadExcel(user_config=config)

    input_data, _ = read_strategy.read(data_path)

    create_res(input_data, args.resfile)


def setup(args):
    """Creates template data"""

    data_type = args.data_type
    data_path = args.data_path
    write_defaults = args.write_defaults
    overwrite = args.overwrite

    if os.path.exists(data_path) and not overwrite:
        raise OtooleSetupError(resource=data_path)

    if data_type == "config":
        shutil.copyfile(
            os.path.join(os.path.dirname(__file__), "preprocess", "config.yaml"),
            data_path,
        )
    elif data_type == "csv":
        config = get_config_setup_data()
        input_data, default_values = get_csv_setup_data(config)
        WriteCsv(user_config=config).write(
            input_data, data_path, default_values, write_defaults=write_defaults
        )


def get_parser():
    parser = argparse.ArgumentParser(
        description="otoole: Python toolkit of OSeMOSYS users"
    )

    parser.add_argument(
        "--verbose", "-v", help="Enable debug mode", action="count", default=0
    )
    parser.add_argument(
        "--version",
        "-V",
        action="version",
        version=__version__,
        help="The version of otoole",
    )

    subparsers = parser.add_subparsers()

    # Parser for results
    result_parser = subparsers.add_parser("results", help="Post-process solution files")
    result_parser.add_argument(
        "from_format",
        help="Result data format to convert from",
        choices=sorted(["cbc", "cplex", "gurobi"]),
    )
    result_parser.add_argument(
        "to_format",
        help="Result data format to convert to",
        choices=sorted(["csv"]),
    )
    result_parser.add_argument(
        "from_path", help="Path to file or folder to convert from"
    )
    result_parser.add_argument("to_path", help="Path to file or folder to convert to")
    result_parser.add_argument(
        "--input_datafile",
        help="Input GNUMathProg datafile required for OSeMOSYS short or fast results",
        default=None,
    )
    result_parser.add_argument(
        "--input_datapackage",
        help="Deprecated",
        default=None,
    )
    result_parser.add_argument("config", help="Path to config YAML file")
    result_parser.add_argument(
        "--write_defaults",
        help="Writes default values",
        default=False,
        action="store_true",
    )
    result_parser.set_defaults(func=result_matrix)

    # Parser for conversion
    convert_parser = subparsers.add_parser(
        "convert", help="Convert from one input format to another"
    )
    convert_parser.add_argument(
        "from_format",
        help="Input data format to convert from",
        choices=sorted(["datafile", "excel", "csv"]),
    )
    convert_parser.add_argument(
        "to_format",
        help="Input data format to convert to",
        choices=sorted(["datafile", "csv", "excel"]),
    )
    convert_parser.add_argument(
        "from_path", help="Path to file or folder to convert from"
    )
    convert_parser.add_argument("to_path", help="Path to file or folder to convert to")
    convert_parser.add_argument("config", help="Path to config YAML file")
    convert_parser.add_argument(
        "--write_defaults",
        help="Writes default values",
        default=False,
        action="store_true",
    )
    convert_parser.add_argument(
        "--keep_whitespace",
        help="Keeps leading/trailing whitespace",
        default=False,
        action="store_true",
    )
    convert_parser.set_defaults(func=conversion_matrix)

    # Parser for validation
    valid_parser = subparsers.add_parser("validate", help="Validate an OSeMOSYS model")
    valid_parser.add_argument(
        "data_format",
        help="Input data format",
        choices=sorted(["datafile", "excel", "csv"]),
    )
    valid_parser.add_argument(
        "data_file", help="Path to the OSeMOSYS data file to validate"
    )
    valid_parser.add_argument("user_config", help="Path to config YAML file")
    valid_parser.add_argument(
        "--validate_config", help="Path to a user-defined validation-config file"
    )
    valid_parser.set_defaults(func=validate_model)

    # Parser for visualisation
    viz_parser = subparsers.add_parser("viz", help="Visualise the model")
    viz_subparsers = viz_parser.add_subparsers()

    res_parser = viz_subparsers.add_parser(
        "res", help="Generate a reference energy system"
    )
    res_parser.add_argument(
        "data_format",
        help="Input data format",
        choices=sorted(["datafile", "excel", "csv"]),
    )
    res_parser.add_argument("data_path", help="Input data path")
    res_parser.add_argument("resfile", help="Path to reference energy system")
    res_parser.add_argument("config", help="Path to config YAML file")
    res_parser.set_defaults(func=data2res)

    # parser for setup
    setup_parser = subparsers.add_parser("setup", help="Setup template files")
    setup_parser.add_argument(
        "data_type", help="Type of file to setup", choices=sorted(["config", "csv"])
    )
    setup_parser.add_argument("data_path", help="Path to file or folder to save to")
    setup_parser.add_argument(
        "--write_defaults",
        help="Writes default values",
        default=False,
        action="store_true",
    )
    setup_parser.add_argument(
        "--overwrite",
        help="Overwrites existing data",
        default=False,
        action="store_true",
    )
    setup_parser.set_defaults(func=setup)

    return parser


class DeprecateAction(argparse.Action):
    def __call__(self, parser, namespace, values, option_string=None):
        logger.warning(f"Argument {self.option_strings} is deprecated and is ignored.")
        delattr(namespace, self.dest)


def main():

    parser = get_parser()
    args = parser.parse_args(sys.argv[1:])

    if args.verbose >= 1 and args.verbose < 2:
        logging.basicConfig(level=logging.INFO)
    if args.verbose > 2:
        logging.basicConfig(level=logging.DEBUG)

    def exception_handler(
        exception_type, exception, traceback, debug_hook=sys.excepthook
    ):
        if args.verbose:
            debug_hook(exception_type, exception, traceback)
        else:
            print("{}: {}".format(exception_type.__name__, str(exception)))

    sys.excepthook = exception_handler

    if "func" in args:
        args.func(args)
    else:
        parser.print_help()



"""Validation methods for the user configuration file."""

import logging
from typing import List, Optional, Union

from pydantic import BaseModel, root_validator, validator

logger = logging.getLogger(__name__)


class UserDefinedValue(BaseModel):
    """Represents any user defined value"""

    name: str
    type: str
    dtype: str
    defined_sets: Optional[List[str]]
    indices: Optional[List[str]]
    default: Optional[Union[int, float]]
    calculated: Optional[bool]
    short_name: Optional[str]

    @validator("type")
    @classmethod
    def check_param_type(cls, value, values):
        if value not in ["param", "result", "set"]:
            raise ValueError(
                f"{values['name']} -> Type must be 'param', 'result', or 'set'"
            )
        return value

    @validator("name", "short_name")
    @classmethod  # for linting purposes
    def check_name_for_spaces(cls, value):
        if " " in value:
            raise ValueError(f"{value} -> Name can not have spaces")
        return value

    @validator("name", "short_name")
    @classmethod
    def check_name_for_numbers(cls, value):
        if any(char.isdigit() for char in value):
            raise ValueError(f"{value} -> Name can not have digits")
        return value

    @validator("name", "short_name")
    @classmethod
    def check_name_for_special_chars(cls, value):
        # removed underscore from the recommeded special char list
        special_characters = " !\"#$%&'()*+,-./:;<=>?@[]^`{|}~"
        if any(c in special_characters for c in value):
            raise ValueError(
                f"{value} -> Name can not have special characters, except for underscores"
            )
        return value

    @root_validator(pre=True)
    @classmethod
    def check_name_length(cls, values):
        if len(values["name"]) > 31:
            if "short_name" not in values:
                raise ValueError(
                    f"{values['name']} -> Name is longer than 31 characters and no 'short_name' field provided"
                )
        if "short_name" in values:
            if len(values["short_name"]) > 31:
                raise ValueError(
                    f"{values['short_name']} -> Name is longer than 31 characters"
                )
        return values

    class Config:
        extra = "forbid"


class UserDefinedSet(UserDefinedValue):
    """Represents a parameter"""

    @validator("dtype")
    @classmethod
    def check_dtype(cls, value, values):
        if value not in ["str", "int"]:
            raise ValueError(f"{values['name']} -> Value must be a 'str' or 'int'")
        return value


class UserDefinedParameter(UserDefinedValue):
    """Represents a user defined parameter"""

    @validator("dtype")
    @classmethod
    def check_dtype(cls, value, values):
        if value not in ["float", "int"]:
            raise ValueError(f"{values['name']} -> Value must be an 'int' or 'float'")
        return value

    @root_validator(pre=True)
    @classmethod
    def check_required_inputs(cls, values):
        required = ["default", "defined_sets", "indices"]
        if not all(req in values for req in required):
            raise ValueError(
                f"{values['name']} -> Missing one of required fields ('default', 'defined_sets', 'indices')"
            )
        return values

    @root_validator(pre=True)
    @classmethod
    def check_index_in_set(cls, values):
        if not all(i in values["defined_sets"] for i in values["indices"]):
            raise ValueError(f"{values['name']} -> Index not in user supplied sets")
        return values

    @root_validator(pre=True)
    @classmethod
    def check_dtype_default(cls, values):
        dtype_input = values["dtype"]
        dtype_default = type(values["default"]).__name__
        if dtype_input != dtype_default:
            # allow ints to be cast as floats
            if not ((dtype_default == "int") and (dtype_input == "float")):
                raise ValueError(
                    f"{values['name']} -> User dtype is {dtype_input} while default value dtype is {dtype_default}"
                )
        return values


class UserDefinedResult(UserDefinedValue):
    """Represents a parameter"""

    @validator("dtype")
    @classmethod
    def check_dtype(cls, value, values):
        if value not in ["float", "int"]:
            raise ValueError(f"{values['name']} -> Value must be an 'int' or 'float'")
        return value

    @root_validator(pre=True)
    @classmethod
    def check_required_inputs(cls, values):
        required = ["default", "defined_sets", "indices", "calculated"]
        if not all(req in values for req in required):
            raise ValueError(
                f"{values['name']} -> Missing one of required fields ('default', 'defined_sets', 'indices', calculated)"
            )
        return values

    @root_validator(pre=True)
    @classmethod
    def check_index_in_set(cls, values):
        if not all(i in values["defined_sets"] for i in values["indices"]):
            raise ValueError(f"{values['name']} -> Index not in user supplied sets")
        return values

    @root_validator(pre=True)
    @classmethod
    def check_dtype_default(cls, values):
        dtype_input = values["dtype"]
        dtype_default = type(values["default"]).__name__
        if dtype_input != dtype_default:
            # allow ints to be cast as floats
            if not ((dtype_default == "int") and (dtype_input == "float")):
                raise ValueError(
                    f"{values['name']} -> User dtype is {dtype_input} while default value dtype is {dtype_default}"
                )
        return values



from tempfile import NamedTemporaryFile

import pandas as pd
import pytest
import yaml

from otoole.exceptions import (
    OtooleDeprecationError,
    OtooleExcelNameLengthError,
    OtooleExcelNameMismatchError,
)
from otoole.read_strategies import ReadExcel
from otoole.utils import (
    UniqueKeyLoader,
    create_name_mappings,
    read_deprecated_datapackage,
)
from otoole.write_strategies import WriteExcel


@pytest.fixture()
def user_config_long_param_name():
    config = {}
    config["REGION"] = {"dtype": "str", "type": "set"}
    config["ParameterNameLongerThanThirtyOneChars"] = {
        "indices": ["REGION"],
        "type": "param",
        "dtype": "float",
        "default": 0.05,
    }
    return config


@pytest.fixture()
def user_config_long_short_name():
    config = {}
    config["REGION"] = {"dtype": "str", "type": "set"}
    config["ParameterNameLongerThanThirtyOneChars"] = {
        "short_name": "ShortNameAlsoLongerThanThirtyOneChars",
        "indices": ["REGION"],
        "type": "param",
        "dtype": "float",
        "default": 0.05,
    }
    return config


class TestCreateNameMappings:
    def test_create_name_mappings(self, user_config):
        expected = {
            "TotalAnnualMaxCapacityInvestment": "TotalAnnualMaxCapacityInvestmen",
            "TotalAnnualMinCapacityInvestment": "TotalAnnualMinCapacityInvestmen",
            "TotalTechnologyAnnualActivityLowerLimit": "TotalTechnologyAnnualActivityLo",
            "TotalTechnologyAnnualActivityUpperLimit": "TotalTechnologyAnnualActivityUp",
            "TotalTechnologyModelPeriodActivityLowerLimit": "TotalTechnologyModelPeriodActLo",
            "TotalTechnologyModelPeriodActivityUpperLimit": "TotalTechnologyModelPeriodActUp",
        }
        actual = create_name_mappings(user_config)
        assert actual == expected

    def test_create_name_mappings_reversed(self, user_config):
        expected = {
            "TotalAnnualMaxCapacityInvestmen": "TotalAnnualMaxCapacityInvestment",
            "TotalAnnualMinCapacityInvestmen": "TotalAnnualMinCapacityInvestment",
            "TotalTechnologyAnnualActivityLo": "TotalTechnologyAnnualActivityLowerLimit",
            "TotalTechnologyAnnualActivityUp": "TotalTechnologyAnnualActivityUpperLimit",
            "TotalTechnologyModelPeriodActLo": "TotalTechnologyModelPeriodActivityLowerLimit",
            "TotalTechnologyModelPeriodActUp": "TotalTechnologyModelPeriodActivityUpperLimit",
        }
        actual = create_name_mappings(user_config, map_full_to_short=False)
        assert actual == expected


def test_excel_name_mismatch_error(user_config):
    read_excel = ReadExcel(user_config=user_config)
    sheet_names = ["AccumulatedAnnualDemand", "MismatchSheet"]
    with pytest.raises(OtooleExcelNameMismatchError):
        read_excel._check_input_sheet_names(sheet_names=sheet_names)


user_config_name_errors = ["user_config_long_param_name", "user_config_long_short_name"]


@pytest.mark.parametrize(
    "user_config_simple",
    user_config_name_errors,
    ids=["full_name_error", "short_name_error"],
)
def test_excel_name_length_error(user_config_simple, request):
    user_config = request.getfixturevalue(user_config_simple)
    write_excel = WriteExcel(user_config=user_config)
    temp_excel = NamedTemporaryFile(suffix=".xlsx")
    handle = pd.ExcelWriter(temp_excel.name)

    with pytest.raises(OtooleExcelNameLengthError):
        write_excel._write_parameter(
            df=pd.DataFrame(),
            parameter_name="ParameterNameLongerThanThirtyOneChars",
            handle=pd.ExcelWriter(handle),
            default=0,
        )


class TestYamlUniqueKeyReader:
    @pytest.fixture()
    def valid_yaml(self):
        data = """
            Key1:
              data1: valid data
              data2: 123
            Key2:
              data1: valid data
              data2: 123
            """
        return data

    invalid_yaml_1 = """
            Key1:
              data1: valid data
              data2: 123
            Key1:
              data1: valid data
              data2: 123
            """

    invalid_yaml_2 = """
            Key1:
              data1: valid data
              data2: 123
            KEY1:
              data1: valid data
              data2: 123
            """

    def test_valid_yaml(self, valid_yaml):
        actual = yaml.load(valid_yaml, Loader=UniqueKeyLoader)
        expected = {
            "Key1": {"data1": "valid data", "data2": 123},
            "Key2": {"data1": "valid data", "data2": 123},
        }
        assert actual == expected

    invalid_yamls = [invalid_yaml_1, invalid_yaml_2]
    invalid_yaml_ids = ["invalid_yaml_1", "invalid_yaml_2"]

    @pytest.mark.parametrize("invalid_yaml", invalid_yamls, ids=invalid_yaml_ids)
    def test_invalid_yaml(self, invalid_yaml):
        with pytest.raises(ValueError):
            yaml.load(invalid_yaml, Loader=UniqueKeyLoader)


def test_successful_read_deprecated_datapackage(tmp_path):
    f = tmp_path / "input/datapackage.json"
    f.parent.mkdir()
    f.touch()
    csvs = tmp_path / "input/data"
    csvs.mkdir()
    actual = read_deprecated_datapackage(f)
    assert actual == str(csvs)


def test_unsuccessful_read_deprecated_datapackage(tmp_path):
    f = tmp_path / "input/datapackage.json"
    f.parent.mkdir()
    f.touch()
    with pytest.raises(OtooleDeprecationError):
        read_deprecated_datapackage(f)



class OtooleException(Exception):
    """Base class for all otoole exceptions."""

    pass


class OtooleValidationError(OtooleException):
    """Input data is invalid

    Arguments
    ---------
    resource : str
        Name of the resource which is invalid
    message : str
        Error message
    """

    def __init__(self, resource, message):
        self.resource = resource
        self.message = message


class OtooleRelationError(OtooleException):
    """Relations between input data is not correct

    Arguments
    ---------
    resource : str
        Name of the resource which is invalid
    foreign_resource : str
        Name of the resource which is invalid
    message : str
        Error message
    """

    def __init__(self, resource, foreign_resource, message):
        self.resource = resource
        self.foreign_resource = foreign_resource
        self.message = message


class OtooleConfigFileError(OtooleException):
    """Config file validation error

    Arguments
    ---------
    message: str
        Message to display to users
    """

    def __init__(self, message: str) -> None:
        self.message = message
        super().__init__(message)

    def __str__(self):
        return f"{self.message}"


class OtooleExcelNameLengthError(OtooleException):
    """Invalid tab name for writing to Excel."""

    def __init__(
        self,
        name: str,
        message: str = "Parameter name must be less than 31 characters when writing to Excel",
    ) -> None:
        self.name = name
        self.message = message
        super().__init__(self.message)

    def __str__(self):
        return f"{self.name} -> {self.message}"


class OtooleExcelNameMismatchError(OtooleException):
    """Name mismatch between config and excel tabs."""

    def __init__(
        self, excel_name: str, message: str = "Excel tab name not found in config file"
    ) -> None:
        self.excel_name = excel_name
        self.message = message
        super().__init__(self.message)

    def __str__(self):
        return f"{self.excel_name} -> {self.message}"


class OtooleDeprecationError(OtooleException):
    """New version of otoole does drops this feature support

    Arguments
    ---------
    resource : str
        Name of the resource which is invalid
    message : str
        Error message
    """

    def __init__(self, resource, message):
        self.resource = resource
        self.message = message

    def __str__(self):
        return f"{self.resource} -> {self.message}"


class OtooleSetupError(OtooleException):
    """Setup data already exists

    Arguments
    ---------
    resource : str
        Name of the resource which is invalid
    message : str
        Error message
    """

    def __init__(
        self,
        resource,
        message="Data already exists. Delete file/directory or pass the --overwrite flag",
    ):
        self.resource = resource
        self.message = message

    def __str__(self):
        return f"{self.resource} -> {self.message}"



import tempfile

import pytest

from otoole.visualise import create_res

url = "tests/fixtures/simplicity-v0.2.1.zip"


@pytest.mark.skip(
    "pydot deprecated, requires replacement with pygraphviz "
    + "(https://github.com/OSeMOSYS/otoole/issues/121)"
)
def test_create_res():

    _, path_to_resfile = tempfile.mkstemp(suffix=".pdf")
    create_res(url, path_to_resfile)



import logging
import os
from typing import Any, TextIO

import pandas as pd

from otoole.exceptions import OtooleExcelNameLengthError
from otoole.input import WriteStrategy

logger = logging.getLogger(__name__)


class WriteExcel(WriteStrategy):
    def _header(self):
        return pd.ExcelWriter(self.filepath, mode="w")

    def _form_parameter(
        self, df: pd.DataFrame, parameter_name: str, default: float
    ) -> pd.DataFrame:
        """Converts data into wide format

        Arguments
        ---------
        df: pd.DataFrame
        parameter_name: str
        default: float

        Returns
        -------
        pandas.DataFrame
        """

        index_names = df.index.names
        column_names = df.columns.to_list()
        if index_names[0]:
            names = index_names + column_names
        else:
            names = column_names
        logger.debug(f"Identified {len(names)} names: {names}")

        total_columns = len(names)

        if total_columns > 3:
            logger.debug("More than 3 columns for {}: {}".format(parameter_name, names))
            rows = names[0:-2]
            columns = names[-2]
            values = names[-1]
            logger.debug(f"Rows: {rows}; columns: {columns}; values: {values}")
            logger.debug("dtypes: {}".format(df.dtypes))
            pivot = df.reset_index().pivot(index=rows, columns=columns, values=values)
        else:
            logger.debug(f"One column for {parameter_name}: {names}")
            pivot = df.copy()

        return pivot

    def _form_parameter_template(self, parameter_name: str, **kwargs) -> pd.DataFrame:
        """Creates wide format excel template

        Arguments
        ---------
        parameter_name: str
        input_data: dict[str, pd.DataFrame])

        Returns
        -------
        pd.DataFrame
        """

        indices = self.user_config[parameter_name]["indices"]

        if "input_data" not in kwargs:
            logger.debug(f"Can not pivot excel template for {parameter_name}")
            return pd.DataFrame(columns=indices)
        else:
            input_data = kwargs["input_data"]

        if "YEAR" in indices:
            years = input_data["YEAR"]["VALUE"].to_list()
            indices.remove("YEAR")
            indices.extend(years)
        else:
            indices.extend(["VALUE"])

        return pd.DataFrame(columns=indices)

    def _write_parameter(
        self,
        df: pd.DataFrame,
        parameter_name: str,
        handle: pd.ExcelWriter,
        default: float,
        **kwargs,
    ):
        try:
            name = self.user_config[parameter_name]["short_name"]
        except KeyError:
            name = parameter_name

        if len(name) > 31:
            raise OtooleExcelNameLengthError(name=name)

        if not df.empty:
            df = self._form_parameter(df, parameter_name, default)
            df.to_excel(handle, sheet_name=name, merge_cells=False, index=True)
        else:
            logger.debug(f"Dataframe {parameter_name} is empty")
            df = self._form_parameter_template(parameter_name, **kwargs)
            df.to_excel(handle, sheet_name=name, merge_cells=False, index=False)

    def _write_set(self, df: pd.DataFrame, set_name, handle: pd.ExcelWriter):
        df.to_excel(handle, sheet_name=set_name, merge_cells=False, index=False)

    def _footer(self, handle=pd.ExcelWriter):
        handle.close()


class WriteDatafile(WriteStrategy):
    def _header(self):
        filepath = open(self.filepath, "w", newline="")
        msg = "# Model file written by *otoole*\n"
        filepath.write(msg)
        return filepath

    def _form_parameter(self, df: pd.DataFrame, default: float):

        # Don't write out values equal to the default value
        df = df[df.VALUE != default]
        return df

    def _write_parameter(
        self,
        df: pd.DataFrame,
        parameter_name: str,
        handle: TextIO,
        default: float,
        **kwargs,
    ):
        """Write parameter data to a GMPL datafile, omitting data with default value

        Arguments
        ---------
        filepath : StreamIO
        df : pandas.DataFrame
        parameter_name : str
        handle: TextIO
        default : int
        """

        if not self.write_defaults:
            df = self._form_parameter(df, default)
        handle.write("param default {} : {} :=\n".format(default, parameter_name))
        df.to_csv(
            path_or_buf=handle, sep=" ", header=False, index=True, float_format="%g"
        )
        handle.write(";\n")

    def _write_set(self, df: pd.DataFrame, set_name, handle: TextIO):
        """Write set data to a GMPL datafile

        Arguments
        ---------
        df : pandas.DataFrame
        set_name : str
        handle: TextIO
        """
        handle.write("set {} :=\n".format(set_name))
        df.to_csv(
            path_or_buf=handle, sep=" ", header=False, index=False, float_format="%g"
        )
        handle.write(";\n")

    def _footer(self, handle: TextIO):
        handle.write("end;\n")
        handle.close()


class WriteCsv(WriteStrategy):
    """Write parameters to comma-separated value files

    Arguments
    ---------
    filepath: str, default=None
        The path to write a folder of csv files
    default_values: dict, default=None
    user_config: dict, default=None
    """

    @staticmethod
    def _write_out_dataframe(folder, parameter, df, index=False):
        """Writes out a dataframe as a csv into a data subfolder

        Arguments
        ---------
        folder : str
        parameter : str
        df : pandas.DataFrame
        index : bool, default=False
            Write the index to CSV
        """
        filepath = os.path.join(folder, parameter + ".csv")
        with open(filepath, "w", newline="") as csvfile:
            logger.info(
                "Writing %s rows into narrow file for %s", df.shape[0], parameter
            )
            df.to_csv(csvfile, index=index)

    def _header(self) -> Any:
        os.makedirs(os.path.join(self.filepath), exist_ok=True)
        return None

    def _write_parameter(
        self,
        df: pd.DataFrame,
        parameter_name: str,
        handle: TextIO,
        default: float,
        **kwargs,
    ) -> pd.DataFrame:
        """Write parameter data"""
        self._write_out_dataframe(self.filepath, parameter_name, df, index=True)

    def _write_set(self, df: pd.DataFrame, set_name, handle: TextIO) -> pd.DataFrame:
        """Write set data"""
        self._write_out_dataframe(self.filepath, set_name, df, index=False)

    def _footer(self, handle: TextIO):
        pass



import logging
from abc import abstractmethod
from io import StringIO
from typing import Any, Dict, List, Set, TextIO, Tuple, Union

import pandas as pd

from otoole.input import ReadStrategy
from otoole.preprocess.longify_data import check_datatypes
from otoole.results.result_package import ResultsPackage

LOGGER = logging.getLogger(__name__)


class ReadResults(ReadStrategy):
    def read(
        self, filepath: Union[str, TextIO], **kwargs
    ) -> Tuple[Dict[str, pd.DataFrame], Dict[str, Any]]:
        """Read a solution file from ``filepath`` and process using ``input_data``

        Arguments
        ---------
        filepath : str, TextIO
            A path name or file buffer pointing to the CBC solution file
        input_data : dict, default=None
            dict of dataframes

        Returns
        -------
        tuple
            A tuple containing dict of pandas.DataFrames and a dict of default_values

        """
        if "input_data" in kwargs:
            input_data = kwargs["input_data"]
        else:
            input_data = None

        available_results = self.get_results_from_file(
            filepath, input_data
        )  # type: Dict[str, pd.DataFrame]

        default_values = self._read_default_values(self.results_config)  # type: Dict

        results = self.calculate_results(
            available_results, input_data
        )  # type: Dict[str, pd.DataFrame]

        return results, default_values

    @abstractmethod
    def get_results_from_file(self, filepath, input_data):
        raise NotImplementedError()

    def calculate_results(
        self,
        available_results: Dict[str, pd.DataFrame],
        input_data: Dict[str, pd.DataFrame],
    ) -> Dict[str, pd.DataFrame]:
        """Populates the results with calculated values using input data"""

        results = {}
        results_package = ResultsPackage(available_results, input_data)

        for name in sorted(self.results_config.keys()):

            LOGGER.info("Looking for %s", name)

            try:
                results[name] = results_package[name]
            except KeyError as ex:
                LOGGER.info("No calculation method available for %s", name)
                LOGGER.debug("Error calculating %s: %s", name, str(ex))

        return results


class ReadResultsCBC(ReadResults):
    def get_results_from_file(self, filepath, input_data):
        cbc = self._convert_to_dataframe(filepath)
        available_results = self._convert_wide_to_long(cbc)
        return available_results

    @abstractmethod
    def _convert_to_dataframe(self, file_path: Union[str, TextIO]) -> pd.DataFrame:
        raise NotImplementedError()

    def _convert_wide_to_long(self, data: pd.DataFrame) -> Dict[str, pd.DataFrame]:
        """Convert from wide to long format

        Converts a pandas DataFrame containing all CBC results to reformatted
        dictionary of pandas DataFrames in long format ready to write out

        Arguments
        ---------
        data : pandas.DataFrame
            CBC results stored in a dataframe

        Example
        -------
        >>> df = pd.DataFrame(data=[
                ['TotalDiscountedCost', "SIMPLICITY,2015", 187.01576],
                ['TotalDiscountedCost', "SIMPLICITY,2016", 183.30788]],
                columns=['Variable', 'Index', 'Value'])
        >>> convert_dataframe_to_csv(df)
        {'TotalDiscountedCost':        REGION  YEAR      VALUE
                                    0  SIMPLICITY  2015  187.01576
                                    1  SIMPLICITY  2016  183.30788}
        """

        sets = {x: y for x, y in self.user_config.items() if y["type"] == "set"}

        results = {}  # type: Dict[str, pd.DataFrame]
        not_found = []

        for name, details in sorted(self.results_config.items()):
            df_cbc = data[data["Variable"] == name]

            if not df_cbc.empty:

                df = df_cbc.copy()  # setting with copy warning
                LOGGER.debug("Extracting results for %s", name)
                indices = details["indices"]  # typing: List

                df[indices] = df["Index"].str.split(",", expand=True)

                types = {index: sets[index]["dtype"] for index in indices}
                df = df.astype(types)

                df = df.drop(columns=["Variable", "Index"])

                df = df.rename(columns={"Value": "VALUE"})

                columns = indices + ["VALUE"]

                df = df[columns]

                index = details["indices"].copy()
                df, index = check_duplicate_index(df, columns, index)
                results[name] = df.set_index(index)
            else:
                not_found.append(name)

        LOGGER.debug("Unable to find result variables for: %s", ", ".join(not_found))

        return results


def check_duplicate_index(df: pd.DataFrame, columns: List, index: List) -> pd.DataFrame:
    """Catches pandas error when there are duplicate column indices"""
    if check_for_duplicates(index):
        index = rename_duplicate_column(index)
        LOGGER.debug("Original column names: %s", columns)
        renamed_columns = rename_duplicate_column(columns)
        LOGGER.debug("New column names: %s", renamed_columns)
        df.columns = renamed_columns
    return df, index


def check_for_duplicates(index: List) -> bool:
    return len(set(index)) != len(index)


def identify_duplicate(index: List) -> Union[int, bool]:
    elements = set()  # type: Set
    for counter, elem in enumerate(index):
        if elem in elements:
            return counter
        else:
            elements.add(elem)
    return False


def rename_duplicate_column(index: List) -> List:
    column = index.copy()
    location = identify_duplicate(column)
    if location:
        column[location] = "_" + column[location]
    return column


class ReadCplex(ReadResults):
    """ """

    def get_results_from_file(
        self, filepath: Union[str, TextIO], input_data
    ) -> Dict[str, pd.DataFrame]:

        if input_data:
            years = input_data["YEAR"].values  # type: List
            start_year = int(years[0])
            end_year = int(years[-1])
        else:
            raise RuntimeError("To process CPLEX results please provide the input file")

        if isinstance(filepath, str):
            with open(filepath, "r") as sol_file:
                data = self.extract_rows(sol_file, start_year, end_year)
        elif isinstance(filepath, StringIO):
            data = self.extract_rows(filepath, start_year, end_year)
        else:
            raise TypeError("Argument filepath type must be a string or an open file")

        results = {}

        for name in data.keys():
            results[name] = self.convert_df(data[name], name, start_year, end_year)

        return results

    def extract_rows(
        self, sol_file: TextIO, start_year: int, end_year: int
    ) -> Dict[str, List[List[str]]]:
        """ """
        data = {}  # type: Dict[str, List[List[str]]]
        for linenum, line in enumerate(sol_file):
            line = line.replace("\n", "")
            try:
                row_as_list = line.split("\t")  # type: List[str]
                name = row_as_list[0]  # type: str

                if name in data.keys():
                    data[name].append(row_as_list)
                else:
                    data[name] = [row_as_list]
            except ValueError as ex:
                msg = "Error caused at line {}: {}. {}"
                raise ValueError(msg.format(linenum, line, ex))
        return data

    def extract_variable_dimensions_values(self, data: List) -> Tuple[str, Tuple, List]:
        """Extracts useful information from a line of a results file"""
        variable = data[0]
        try:
            number = len(self.results_config[variable]["indices"])
        except KeyError as ex:
            print(data)
            raise KeyError(ex)
        dimensions = tuple(data[1:(number)])
        values = data[(number):]
        return (variable, dimensions, values)

    def convert_df(
        self, data: List[List[str]], variable: str, start_year: int, end_year: int
    ) -> pd.DataFrame:
        """Read the cplex lines into a pandas DataFrame"""
        index = self.results_config[variable]["indices"]
        columns = ["variable"] + index[:-1] + list(range(start_year, end_year + 1, 1))
        df = pd.DataFrame(data=data, columns=columns)
        df, index = check_duplicate_index(df, columns, index)
        df = df.drop(columns="variable")

        LOGGER.debug(
            f"Attempting to set index for {variable} with columns {index[:-1]}"
        )
        try:
            df = df.set_index(index[:-1])
        except NotImplementedError as ex:
            LOGGER.error(f"Error setting index for {df.head()}")
            raise NotImplementedError(ex)
        df = df.melt(var_name="YEAR", value_name="VALUE", ignore_index=False)
        df = df.reset_index()
        df = check_datatypes(df, self.user_config, variable)
        df = df.set_index(index)
        df = df[(df != 0).any(axis=1)]
        return df


class ReadGurobi(ReadResultsCBC):
    """Read a Gurobi solution file into memory"""

    def _convert_to_dataframe(self, file_path: Union[str, TextIO]) -> pd.DataFrame:
        """Reads a Gurobi solution file into a pandas DataFrame

        Arguments
        ---------
        file_path : str
        """
        df = pd.read_csv(
            file_path,
            header=None,
            sep=" ",
            names=["Variable", "Value"],
            skiprows=2,
        )  # type: pd.DataFrame
        df[["Variable", "Index"]] = df["Variable"].str.split("(", expand=True)
        df["Index"] = df["Index"].str.replace(")", "", regex=True)
        LOGGER.debug(df)
        df = df[(df["Value"] != 0)].reset_index()
        return df[["Variable", "Index", "Value"]].astype({"Value": float})


class ReadCbc(ReadResultsCBC):
    """Read a CBC solution file into memory

    Arguments
    ---------
    user_config
    results_config
    """

    def _convert_to_dataframe(self, file_path: Union[str, TextIO]) -> pd.DataFrame:
        """Reads a CBC solution file into a pandas DataFrame

        Arguments
        ---------
        file_path : str
        """
        df = pd.read_csv(
            file_path,
            header=None,
            sep="(",
            names=["Variable", "indexvalue"],
            skiprows=1,
        )  # type: pd.DataFrame
        if df["Variable"].astype(str).str.contains(r"^\*\*").any():
            LOGGER.warning(
                "CBC Solution File contains decision variables out of bounds. "
                + "You have an infeasible solution"
            )
        df["Variable"] = (
            df["Variable"]
            .astype(str)
            .str.replace(r"^\*\*", "", regex=True)
            .str.split(expand=True)[1]
        )
        df[["Index", "Value"]] = df["indexvalue"].str.split(expand=True).loc[:, 0:1]
        df["Index"] = df["Index"].str.replace(")", "", regex=True)
        df = df.drop(columns=["indexvalue"])
        return df[["Variable", "Index", "Value"]].astype({"Value": float})



import pandas as pd
from pandas.testing import assert_frame_equal

from otoole.preprocess.setup import get_csv_setup_data


def test_get_csv_setup_data():
    config = {
        "AccumulatedAnnualDemand": {
            "indices": ["REGION", "FUEL", "YEAR"],
            "type": "param",
            "dtype": "float",
            "default": 0,
        },
        "REGION": {
            "dtype": "str",
            "type": "set",
        },
        "FUEL": {
            "dtype": "str",
            "type": "set",
        },
        "YEAR": {
            "dtype": "int",
            "type": "set",
        },
        "Demand": {
            "indices": ["REGION", "TIMESLICE", "FUEL", "YEAR"],
            "type": "result",
            "dtype": "float",
            "default": 0,
            "calculated": True,
        },
    }

    accumulated_annual_demand = pd.DataFrame(
        columns=["REGION", "FUEL", "YEAR", "VALUE"]
    ).astype(
        {
            "REGION": "str",
            "FUEL": "str",
            "YEAR": "int64",
            "VALUE": "float",
        }
    )
    expected_inputs = {
        "AccumulatedAnnualDemand": accumulated_annual_demand.set_index(
            ["REGION", "FUEL", "YEAR"]
        ),
        "REGION": pd.DataFrame(columns=["VALUE"]),
        "FUEL": pd.DataFrame(columns=["VALUE"]),
        "YEAR": pd.DataFrame(columns=["VALUE"]),
    }
    expected_defaults = {
        "AccumulatedAnnualDemand": 0,
        "Demand": 0,
    }
    actual_inputs, actual_defaults = get_csv_setup_data(config)

    assert_frame_equal(
        actual_inputs["AccumulatedAnnualDemand"],
        expected_inputs["AccumulatedAnnualDemand"],
    )
    assert_frame_equal(actual_inputs["REGION"], expected_inputs["REGION"])
    assert "Demand" not in actual_inputs
    assert actual_defaults == expected_defaults



import json
import logging
import os
from typing import Any, Dict, List, Union

import pandas as pd
from pydantic import ValidationError
from yaml import SafeLoader, load  # type: ignore

from otoole.exceptions import OtooleConfigFileError, OtooleDeprecationError
from otoole.preprocess.validate_config import (
    UserDefinedParameter,
    UserDefinedResult,
    UserDefinedSet,
    UserDefinedValue,
)

try:
    import importlib.resources as resources
except ImportError:
    # Try backported to PY<37 `importlib_resources`.
    import importlib_resources as resources  # type: ignore

logger = logging.getLogger(__name__)


def _read_file(open_file, ending):
    if ending == ".yaml" or ending == ".yml":
        contents = load(open_file, Loader=UniqueKeyLoader)  # typing: Dict
    elif ending == ".json":
        contents = json.load(open_file)  # typing: Dict
    else:
        contents = open_file.readlines()
    return contents


def read_packaged_file(filename: str, module_name: str = None):

    _, ending = os.path.splitext(filename)

    if module_name is None:
        with open(filename, "r") as open_file:
            contents = _read_file(open_file, ending)
    else:
        with resources.open_text(module_name, filename) as open_file:
            contents = _read_file(open_file, ending)

    return contents


def extract_config(
    schema: Dict, default_values: Dict
) -> Dict[str, Dict[str, Union[str, List[str]]]]:

    config = {}  # type: Dict[str, Dict[str, Union[str, List[str]]]]
    for resource in schema["resources"]:

        name = resource["name"]
        if name == "default_values":
            continue

        dtype_mapping = {
            "number": "float",
            "string": "str",
            "float": "float",
            "integer": "int",
        }

        fields = resource["schema"]["fields"]
        dtype = [x["type"] for x in fields if x["name"] == "VALUE"][0]
        if (len(fields) == 1) & (fields[0]["name"] == "VALUE"):
            element_type = "set"
            config[name] = {"dtype": dtype_mapping[dtype], "type": element_type}
        else:
            element_type = "param"
            indices = [x["name"] for x in fields if x["name"] != "VALUE"]
            config[name] = {
                "type": element_type,
                "indices": indices,
                "dtype": dtype_mapping[dtype],
                "default": default_values[name],
            }
    return config


def create_name_mappings(
    config: Dict[str, Dict[str, Union[str, List]]], map_full_to_short: bool = True
) -> Dict:
    """Creates name mapping between full name and short name.

    Arguments
    ---------
    config : Dict[str, Dict[str, Union[str, List]]]
        Parsed user configuration file
    map_full_to_short: bool
        Map full name to short name if true, else map short name to full name

    Returns
    -------
    csv_to_excel Dict[str, str]
        Mapping of full name to shortened name

    """

    csv_to_excel = {}
    for name, params in config.items():
        try:
            csv_to_excel[name] = params["short_name"]
        except KeyError:
            if len(name) > 31:
                logger.info(f"{name} does not have a 'short_name'")
            continue

    if map_full_to_short:
        return csv_to_excel
    else:
        return {v: k for k, v in csv_to_excel.items()}


def validate_config(config: Dict) -> None:
    """Validates user input data

    Arguments
    ---------
    config: Dict
        Read in user config yaml file

    Raises
    ------
    ValidationError
        If the user inputs are not valid
    """

    # For validating with pydantic
    config_flattened = format_config_for_validation(config)
    user_defined_sets = get_all_sets(config)

    errors = []
    for input_data in config_flattened:
        try:
            if "type" not in input_data:
                UserDefinedValue(**input_data)
            elif input_data["type"] == "param":
                input_data["defined_sets"] = user_defined_sets
                UserDefinedParameter(**input_data)
            elif input_data["type"] == "result":
                input_data["defined_sets"] = user_defined_sets
                UserDefinedResult(**input_data)
            elif input_data["type"] == "set":
                UserDefinedSet(**input_data)
            else:
                # have pydantic raise an error
                UserDefinedValue(
                    name=input_data["name"],
                    type=input_data["type"],
                    dtype=input_data["dtype"],
                )
        except ValidationError as ex:
            errors_caught = [x["msg"] for x in ex.errors()]
            errors.extend(errors_caught)

    if errors:
        error_message = "\n".join(errors)
        raise OtooleConfigFileError(message=f"\n{error_message}")


def format_config_for_validation(config_in: Dict) -> List:
    """Formats config for validation function.

    Flattens dictionary to a list

    Arguments
    ---------
    config_in: Dict
        Read in user config yaml file

    Returns
    -------
    config_out: List

    Example
    -------
    >>> config_in
    >>> AccumulatedAnnualDemand:
          indices: [REGION,FUEL,YEAR]
          type: param
          dtype: float
          default: 0

    >>> config_out
    >>> [{
        name: AccumulatedAnnualDemand
        indices: [REGION,FUEL,YEAR]
        type: param
        dtype: float
        default: 0
        }, ... ]
    """
    config_out = []
    for name, data in config_in.items():
        flattened_data = {"name": name, **data}
        config_out.append(flattened_data)
    return config_out


def read_deprecated_datapackage(datapackage: str) -> str:
    """Checks filepath for CSVs if a datapackage is provided

    Arguments
    ---------
    datapackage: str
        Location of input datapackge

    Returns
    -------
    input_csvs: str
        Location of input csv files

    Raises
    ------
    OtooleDeprecationError
        If no 'data/' directory is found
    """

    input_csvs = os.path.join(os.path.dirname(datapackage), "data")
    if os.path.exists(input_csvs):
        return input_csvs
    else:
        raise OtooleDeprecationError(
            resource="datapackage.json",
            message="datapackage format no longer supported and no csv data found",
        )


def get_packaged_resource(
    input_data: Dict[str, pd.DataFrame], param: str
) -> List[Dict[str, Any]]:
    """Gets input parameter data and formats it as a dictionary

    Arguments
    ---------
    input_data : Dict[str, pd.DataFrame]
        Internal datastore for otoole input data
    param : str
        Name of OSeMOSYS parameter

    Returns
    -------
    List[Dict[str,any]]
        List of all rows in the df, where each dictionary is the column
        name, followed by the value in that row

    Example
    -------
    >>> get_packaged_resource(input_data, "InputActivityRatio")
    >>> [{'REGION': 'SIMPLICITY',
        'TECHNOLOGY': 'RIVWATAGR',
        'FUEL': 'WATIN',
        'MODE_OF_OPERATION': 1,
        'YEAR': 2020,
        'VALUE': 1.0}]
    """
    return input_data[param].reset_index().to_dict(orient="records")


class UniqueKeyLoader(SafeLoader):
    """YALM Loader to find duplicate keys

    This loader will treat lowercase and uppercase keys as the same. Meaning,
    the keys "SampleKey" and "SAMPLEKEY" are considered the same.

    Raises
    ------
    ValueError
        If a key is defined more than once.

    Adapted from:
    https://stackoverflow.com/a/63215043/14961492
    """

    def construct_mapping(self, node, deep=False):
        mapping = []
        for key_node, _ in node.value:
            key = self.construct_object(key_node, deep=deep)
            key = key.upper()
            if key in mapping:
                raise ValueError(f"{key} -> defined more than once")
            mapping.append(key)
        return super().construct_mapping(node, deep)


def get_all_sets(config: Dict) -> List:
    """Extracts user defined sets"""
    return [x for x, y in config.items() if y["type"] == "set"]



import os
from io import StringIO
from textwrap import dedent
from typing import List

import pandas as pd
from amply import Amply
from pytest import mark, raises

from otoole import ReadCsv, ReadDatafile, ReadExcel, ReadMemory
from otoole.exceptions import OtooleDeprecationError
from otoole.preprocess.longify_data import check_datatypes
from otoole.results.results import (
    ReadCbc,
    ReadCplex,
    ReadGurobi,
    check_for_duplicates,
    identify_duplicate,
    rename_duplicate_column,
)


class TestReadCplex:

    cplex_empty = (
        "AnnualFixedOperatingCost	REGION	AOBACKSTOP	0	0	0	0	0	0	0	0	0	0	0	0	0	0	0	0"
    )
    cplex_short = "AnnualFixedOperatingCost	REGION	CDBACKSTOP	0.0	0.0	137958.8400384134	305945.38410619126	626159.9611543404	0.0	0.0	0.0	0.0	0.0	0.0	0.0	0.0	0.0	0.0	0.0"
    cplex_long = "RateOfActivity	REGION	S1D1	CGLFRCFURX	1	0.0	0.0	0.0	0.0	0.0	0.3284446367303371	0.3451714779880536	0.3366163200621617	0.3394945166233896	0.3137488154250392	0.28605725055560716	0.2572505015401749	0.06757558148965725	0.0558936625751148	0.04330608461292407	0.0"

    cplex_mid_empty = (
        pd.DataFrame(
            data=[],
            columns=["REGION", "TECHNOLOGY", "YEAR", "VALUE"],
        )
        .astype({"VALUE": float})
        .set_index(["REGION", "TECHNOLOGY", "YEAR"])
    )

    cplex_mid_short = pd.DataFrame(
        data=[
            ["REGION", "CDBACKSTOP", 2017, 137958.8400384134],
            ["REGION", "CDBACKSTOP", 2018, 305945.38410619126],
            ["REGION", "CDBACKSTOP", 2019, 626159.9611543404],
        ],
        columns=["REGION", "TECHNOLOGY", "YEAR", "VALUE"],
    ).set_index(["REGION", "TECHNOLOGY", "YEAR"])

    cplex_mid_long = pd.DataFrame(
        data=[
            ["REGION", "S1D1", "CGLFRCFURX", 1, 2020, 0.3284446367303371],
            ["REGION", "S1D1", "CGLFRCFURX", 1, 2021, 0.3451714779880536],
            ["REGION", "S1D1", "CGLFRCFURX", 1, 2022, 0.3366163200621617],
            ["REGION", "S1D1", "CGLFRCFURX", 1, 2023, 0.3394945166233896],
            ["REGION", "S1D1", "CGLFRCFURX", 1, 2024, 0.3137488154250392],
            ["REGION", "S1D1", "CGLFRCFURX", 1, 2025, 0.28605725055560716],
            ["REGION", "S1D1", "CGLFRCFURX", 1, 2026, 0.2572505015401749],
            ["REGION", "S1D1", "CGLFRCFURX", 1, 2027, 0.06757558148965725],
            ["REGION", "S1D1", "CGLFRCFURX", 1, 2028, 0.0558936625751148],
            ["REGION", "S1D1", "CGLFRCFURX", 1, 2029, 0.04330608461292407],
        ],
        columns=[
            "REGION",
            "TIMESLICE",
            "TECHNOLOGY",
            "MODE_OF_OPERATION",
            "YEAR",
            "VALUE",
        ],
    ).set_index(["REGION", "TIMESLICE", "TECHNOLOGY", "MODE_OF_OPERATION", "YEAR"])

    dataframe_short = {
        "AnnualFixedOperatingCost": pd.DataFrame(
            data=[
                ["REGION", "CDBACKSTOP", 2017, 137958.8400384134],
                ["REGION", "CDBACKSTOP", 2018, 305945.3841061913],
                ["REGION", "CDBACKSTOP", 2019, 626159.9611543404],
            ],
            columns=["REGION", "TECHNOLOGY", "YEAR", "VALUE"],
        ).set_index(["REGION", "TECHNOLOGY", "YEAR"])
    }

    dataframe_long = {
        "RateOfActivity": pd.DataFrame(
            data=[
                ["REGION", "S1D1", "CGLFRCFURX", 1, 2020, 0.3284446367303371],
                ["REGION", "S1D1", "CGLFRCFURX", 1, 2021, 0.3451714779880536],
                ["REGION", "S1D1", "CGLFRCFURX", 1, 2022, 0.3366163200621617],
                ["REGION", "S1D1", "CGLFRCFURX", 1, 2023, 0.3394945166233896],
                ["REGION", "S1D1", "CGLFRCFURX", 1, 2024, 0.3137488154250392],
                ["REGION", "S1D1", "CGLFRCFURX", 1, 2025, 0.28605725055560716],
                ["REGION", "S1D1", "CGLFRCFURX", 1, 2026, 0.2572505015401749],
                ["REGION", "S1D1", "CGLFRCFURX", 1, 2027, 0.06757558148965725],
                ["REGION", "S1D1", "CGLFRCFURX", 1, 2028, 0.0558936625751148],
                ["REGION", "S1D1", "CGLFRCFURX", 1, 2029, 0.04330608461292407],
            ],
            columns=[
                "REGION",
                "TIMESLICE",
                "TECHNOLOGY",
                "MODE_OF_OPERATION",
                "YEAR",
                "VALUE",
            ],
        ).set_index(["REGION", "TIMESLICE", "TECHNOLOGY", "MODE_OF_OPERATION", "YEAR"])
    }

    test_data = [
        (cplex_short, dataframe_short),
        (cplex_long, dataframe_long),
    ]

    @mark.parametrize("cplex_input,expected", test_data, ids=["short", "long"])
    def test_read_cplex_to_dataframe(self, cplex_input, expected, user_config):
        cplex_reader = ReadCplex(user_config=user_config)

        input_data = {
            "YEAR": pd.DataFrame(data=list(range(2015, 2031, 1)), columns=["VALUE"]),
            "REGION": pd.DataFrame(data=["REGION"], columns=["VALUE"]),
            "TECHNOLOGY": pd.DataFrame(
                data=["CDBACKSTOP", "CGLFRCFURX"], columns=["VALUE"]
            ),
            "MODE_OF_OPERATION": pd.DataFrame(data=[1], columns=["VALUE"]),
            "TIMESLICE": pd.DataFrame(data=["S1D1"], columns=["VALUE"]),
        }

        with StringIO(cplex_input) as file_buffer:
            actual, _ = cplex_reader.read(file_buffer, input_data=input_data)
        for name, item in actual.items():
            pd.testing.assert_frame_equal(item, expected[name])

    test_data_mid = [(cplex_short, cplex_mid_short), (cplex_long, cplex_mid_long)]

    def test_read_empty_cplex_to_dataframe(self, user_config):
        cplex_input = self.cplex_empty

        cplex_reader = ReadCplex(user_config)

        input_data = {
            "YEAR": pd.DataFrame(data=list(range(2015, 2031, 1)), columns=["VALUE"])
        }

        with StringIO(cplex_input) as file_buffer:
            data, _ = cplex_reader.read(file_buffer, input_data=input_data)
        assert "AnnualFixedOperatingCost" in data
        expected = (
            pd.DataFrame(
                data=[],
                columns=["REGION", "TECHNOLOGY", "YEAR", "VALUE"],
            )
            .astype({"REGION": str, "VALUE": float, "YEAR": int, "TECHNOLOGY": str})
            .set_index(["REGION", "TECHNOLOGY", "YEAR"])
        )
        actual = data["AnnualFixedOperatingCost"]
        pd.testing.assert_frame_equal(actual, expected, check_index_type=False)

    test_data_to_cplex = [
        (cplex_empty, cplex_mid_empty),
        (cplex_short, cplex_mid_short),
        (cplex_long, cplex_mid_long),
    ]

    @mark.parametrize(
        "cplex_input,expected", test_data_to_cplex, ids=["empty", "short", "long"]
    )
    def test_convert_cplex_to_df(self, cplex_input, expected, user_config):

        data = cplex_input.split("\t")
        variable = data[0]
        cplex_reader = ReadCplex(user_config=user_config)
        actual = cplex_reader.convert_df([data], variable, 2015, 2030)
        pd.testing.assert_frame_equal(actual, expected, check_index_type=False)

    def test_convert_lines_to_df_empty(self, user_config):

        data = [
            [
                "AnnualFixedOperatingCost",
                "REGION",
                "AOBACKSTOP",
                "0",
                "0",
                "0",
                "0",
                "0",
                "0",
                "0",
                "0",
                "0",
            ]
        ]
        variable = "AnnualFixedOperatingCost"
        cplex_reader = ReadCplex(user_config)
        actual = cplex_reader.convert_df(data, variable, 2015, 2023)
        pd.testing.assert_frame_equal(
            actual,
            pd.DataFrame(
                data=[],
                columns=["REGION", "TECHNOLOGY", "YEAR", "VALUE"],
            )
            .astype({"REGION": str, "TECHNOLOGY": str, "YEAR": int, "VALUE": float})
            .set_index(["REGION", "TECHNOLOGY", "YEAR"]),
            check_index_type=False,
        )

    def test_check_datatypes_with_empty(self):

        df = pd.DataFrame(data=[], columns=["REGION", "FUEL", "YEAR", "VALUE"])

        parameter = "AccumulatedAnnualDemand"

        config_dict = {
            "AccumulatedAnnualDemand": {
                "indices": ["REGION", "FUEL", "YEAR"],
                "type": "param",
                "dtype": float,
                "default": 0,
            },
            "REGION": {"dtype": "str", "type": "set"},
            "FUEL": {"dtype": "str", "type": "set"},
            "YEAR": {"dtype": "int", "type": "set"},
        }

        actual = check_datatypes(df, config_dict, parameter)

        expected = pd.DataFrame(
            data=[], columns=["REGION", "FUEL", "YEAR", "VALUE"]
        ).astype({"REGION": str, "FUEL": str, "YEAR": int, "VALUE": float})

        pd.testing.assert_frame_equal(actual, expected, check_index_type=False)


class TestReadGurobi:

    gurobi_data = dedent(
        """# Solution for model cost
# Objective value = 4.4973196701520455e+03
TotalDiscountedCost(SIMPLICITY,2013) 0
TotalDiscountedCost(SIMPLICITY,2014) 1.9360385416218188e+02
TotalDiscountedCost(SIMPLICITY,2015) 1.8772386050936669e+02
TotalDiscountedCost(SIMPLICITY,2016) 1.8399762956864294e+02
TotalDiscountedCost(SIMPLICITY,2017) 1.8172752298186381e+02
RateOfActivity(SIMPLICITY,ID,FEL1,1,2014) 1.59376124775045
RateOfActivity(SIMPLICITY,ID,FEL1,1,2015) 1.60167966406719
RateOfActivity(SIMPLICITY,ID,FEL1,1,2016) 1.6369526094781
RateOfActivity(SIMPLICITY,ID,FEL1,1,2017) 1.68590281943611
"""
    )

    def test_convert_to_dataframe(self, user_config):
        input_file = self.gurobi_data
        reader = ReadGurobi(user_config)
        with StringIO(input_file) as file_buffer:
            actual = reader._convert_to_dataframe(file_buffer)
        # print(actual)
        expected = pd.DataFrame(
            [
                ["TotalDiscountedCost", "SIMPLICITY,2014", 1.9360385416218188e02],
                ["TotalDiscountedCost", "SIMPLICITY,2015", 1.8772386050936669e02],
                ["TotalDiscountedCost", "SIMPLICITY,2016", 1.8399762956864294e02],
                ["TotalDiscountedCost", "SIMPLICITY,2017", 1.8172752298186381e02],
                ["RateOfActivity", "SIMPLICITY,ID,FEL1,1,2014", 1.59376124775045],
                ["RateOfActivity", "SIMPLICITY,ID,FEL1,1,2015", 1.60167966406719],
                ["RateOfActivity", "SIMPLICITY,ID,FEL1,1,2016", 1.6369526094781],
                ["RateOfActivity", "SIMPLICITY,ID,FEL1,1,2017", 1.68590281943611],
            ],
            columns=["Variable", "Index", "Value"],
        ).astype({"Variable": str, "Index": str, "Value": float})

        pd.testing.assert_frame_equal(actual, expected)

    def test_solution_to_dataframe(self, user_config):
        input_file = self.gurobi_data
        reader = ReadGurobi(user_config)
        with StringIO(input_file) as file_buffer:
            actual = reader.read(file_buffer)
        # print(actual)
        expected = (
            pd.DataFrame(
                [
                    ["SIMPLICITY", 2014, 1.9360385416218188e02],
                    ["SIMPLICITY", 2015, 1.8772386050936669e02],
                    ["SIMPLICITY", 2016, 1.8399762956864294e02],
                    ["SIMPLICITY", 2017, 1.8172752298186381e02],
                ],
                columns=["REGION", "YEAR", "VALUE"],
            )
            .astype({"YEAR": int, "VALUE": float})
            .set_index(["REGION", "YEAR"])
        )

        pd.testing.assert_frame_equal(actual[0]["TotalDiscountedCost"], expected)

        expected = (
            pd.DataFrame(
                [
                    ["SIMPLICITY", "ID", "FEL1", 1, 2014, 1.59376124775045],
                    ["SIMPLICITY", "ID", "FEL1", 1, 2015, 1.60167966406719],
                    ["SIMPLICITY", "ID", "FEL1", 1, 2016, 1.6369526094781],
                    ["SIMPLICITY", "ID", "FEL1", 1, 2017, 1.68590281943611],
                ],
                columns=[
                    "REGION",
                    "TIMESLICE",
                    "TECHNOLOGY",
                    "MODE_OF_OPERATION",
                    "YEAR",
                    "VALUE",
                ],
            )
            .astype({"YEAR": int, "VALUE": float, "MODE_OF_OPERATION": int})
            .set_index(
                ["REGION", "TIMESLICE", "TECHNOLOGY", "MODE_OF_OPERATION", "YEAR"]
            )
        )
        pd.testing.assert_frame_equal(actual[0]["RateOfActivity"], expected)


class TestReadCbc:

    cbc_data = dedent(
        """0 Trade(Globe,Globe,IP,L_AGR,2015) -0.0 0
0 Trade(Globe,Globe,IP,L_AGR,2016) -1.0 0
0 Trade(Globe,Globe,IP,L_AGR,2017) -2.0 0
0 Trade(Globe,Globe,IP,L_AGR,2018) -3.0 0
0 Trade(Globe,Globe,IP,L_AGR,2019) -4.0 0
0 Trade(Globe,Globe,IP,L_AGR,2020) -5.0 0
            """
    )
    otoole_data = pd.DataFrame(
        data=[
            ["Globe", "Globe", "IP", "L_AGR", 2016, -1.0],
            ["Globe", "Globe", "IP", "L_AGR", 2017, -2.0],
            ["Globe", "Globe", "IP", "L_AGR", 2018, -3.0],
            ["Globe", "Globe", "IP", "L_AGR", 2019, -4.0],
            ["Globe", "Globe", "IP", "L_AGR", 2020, -5.0],
        ],
        columns=["REGION", "_REGION", "TIMESLICE", "FUEL", "YEAR", "VALUE"],
    ).set_index(["REGION", "_REGION", "TIMESLICE", "FUEL", "YEAR"])

    test_data = [
        (
            cbc_data,
            otoole_data,
        )
    ]

    @mark.parametrize("cbc_input,expected", test_data)
    def test_read_cbc_to_otoole_dataframe(self, cbc_input, expected, user_config):
        with StringIO(cbc_input) as file_buffer:
            actual = ReadCbc(user_config).read(file_buffer, kwargs={"input_data": {}})[
                0
            ]["Trade"]
        pd.testing.assert_frame_equal(actual, expected)

    def test_read_cbc_dataframe_to_otoole_dataframe(self, user_config):

        prelim_data = pd.DataFrame(
            data=[
                ["Trade", "Globe,Globe,IP,L_AGR,2016", -1.0],
                ["Trade", "Globe,Globe,IP,L_AGR,2017", -2.0],
                ["Trade", "Globe,Globe,IP,L_AGR,2018", -3.0],
                ["Trade", "Globe,Globe,IP,L_AGR,2019", -4.0],
                ["Trade", "Globe,Globe,IP,L_AGR,2020", -5.0],
            ],
            columns=["Variable", "Index", "Value"],
        )
        actual = ReadCbc(user_config)._convert_wide_to_long(prelim_data)["Trade"]
        pd.testing.assert_frame_equal(actual, self.otoole_data)

    test_data_4 = [
        (["REGION", "REGION", "TIMESLICE", "FUEL", "YEAR"], True),
        (["REGION", "TIMESLICE", "FUEL", "YEAR"], False),
        ([], False),
    ]

    @mark.parametrize("data,expected", test_data_4)
    def test_handle_duplicate_indices(self, data, expected):
        assert check_for_duplicates(data) is expected

    test_data_5 = [
        (["REGION", "REGION", "TIMESLICE", "FUEL", "YEAR"], 1),
        (["REGION", "TIMESLICE", "FUEL", "YEAR"], False),
        ([], False),
    ]

    @mark.parametrize("data,expected", test_data_5)
    def test_identify_duplicate(self, data, expected):
        assert identify_duplicate(data) == expected

    def test_rename_duplicate_column(self):
        data = ["REGION", "REGION", "TIMESLICE", "FUEL", "YEAR"]
        actual = rename_duplicate_column(data)
        expected = ["REGION", "_REGION", "TIMESLICE", "FUEL", "YEAR"]
        assert actual == expected

    total_cost_cbc = dedent(
        """Optimal - objective value 4483.96932237
                             1 TotalDiscountedCost(SIMPLICITY,2015){0}187.01576{1}0
                             2 TotalDiscountedCost(SIMPLICITY,2016){0}183.30788{1}0
                             3 TotalDiscountedCost(SIMPLICITY,2017){0}181.05465{1}0
                             4 TotalDiscountedCost(SIMPLICITY,2018){0}218.08923{1}0
                             5 TotalDiscountedCost(SIMPLICITY,2019){0}193.85792{1}0
                             6 TotalDiscountedCost(SIMPLICITY,2020){0}233.79202{1}0

                         """.format(
            " " * 51, " " * 23
        )
    )

    total_cost_cbc_mid = pd.DataFrame(
        data=[
            ["TotalDiscountedCost", "SIMPLICITY,2015", 187.01576],
            ["TotalDiscountedCost", "SIMPLICITY,2016", 183.30788],
            ["TotalDiscountedCost", "SIMPLICITY,2017", 181.05465],
            ["TotalDiscountedCost", "SIMPLICITY,2018", 218.08923],
            ["TotalDiscountedCost", "SIMPLICITY,2019", 193.85792],
            ["TotalDiscountedCost", "SIMPLICITY,2020", 233.79202],
        ],
        columns=["Variable", "Index", "Value"],
    )

    total_cost_otoole_df = {
        "TotalDiscountedCost": pd.DataFrame(
            data=[
                ["SIMPLICITY", 2015, 187.01576],
                ["SIMPLICITY", 2016, 183.30788],
                ["SIMPLICITY", 2017, 181.05465],
                ["SIMPLICITY", 2018, 218.08923],
                ["SIMPLICITY", 2019, 193.85792],
                ["SIMPLICITY", 2020, 233.79202],
            ],
            columns=["REGION", "YEAR", "VALUE"],
        ).set_index(["REGION", "YEAR"])
    }

    test_data = [(total_cost_cbc, total_cost_cbc_mid)]

    @mark.parametrize("cbc_input,expected", test_data, ids=["TotalDiscountedCost"])
    def test_read_cbc_to_dataframe(self, cbc_input, expected, user_config):
        cbc_reader = ReadCbc(user_config)
        with StringIO(cbc_input) as file_buffer:
            actual = cbc_reader._convert_to_dataframe(file_buffer)
        pd.testing.assert_frame_equal(actual, expected)

    test_data_2 = [
        # First case
        (total_cost_cbc_mid, total_cost_otoole_df),
        # Second case
        (
            pd.DataFrame(
                data=[
                    ["AnnualEmissions", "REGION,CO2,2017", 137958.8400384134],
                    ["AnnualEmissions", "REGION,CO2,2018", 305945.3841061913],
                    ["AnnualEmissions", "REGION,CO2,2019", 626159.9611543404],
                ],
                columns=["Variable", "Index", "Value"],
            ),
            {
                "AnnualEmissions": pd.DataFrame(
                    data=[
                        ["REGION", "CO2", 2017, 137958.8400384134],
                        ["REGION", "CO2", 2018, 305945.3841061913],
                        ["REGION", "CO2", 2019, 626159.9611543404],
                    ],
                    columns=["REGION", "EMISSION", "YEAR", "VALUE"],
                ).set_index(["REGION", "EMISSION", "YEAR"])
            },
        ),
    ]  # type: List

    @mark.parametrize(
        "results,expected",
        test_data_2,
        ids=["TotalDiscountedCost", "AnnualEmissions1"],
    )
    def test_convert_cbc_to_csv_long(self, results, expected, user_config):
        cbc_reader = ReadCbc(user_config=user_config)
        actual = cbc_reader._convert_wide_to_long(results)
        assert isinstance(actual, dict)
        for name, df in actual.items():
            pd.testing.assert_frame_equal(df, expected[name])

    test_data_3 = [(total_cost_cbc, {}, total_cost_otoole_df)]  # type: List

    @mark.parametrize(
        "cbc_solution,input_data,expected",
        test_data_3,
        ids=["TotalDiscountedCost"],
    )
    def test_convert_cbc_to_csv_long_read(
        self, cbc_solution, input_data, expected, user_config
    ):
        cbc_reader = ReadCbc(user_config=user_config)
        with StringIO(cbc_solution) as file_buffer:
            actual = cbc_reader.read(file_buffer, kwargs={"input_data": input_data})[0][
                "TotalDiscountedCost"
            ]
        assert isinstance(actual, pd.DataFrame)
        pd.testing.assert_frame_equal(actual, expected["TotalDiscountedCost"])

    def test_calculate_results(self, user_config):
        cbc_results = {
            "RateOfActivity": pd.DataFrame(
                data=[
                    ["SIMPLICITY", "GAS_EXTRACTION", "ID", 1, 2014, 1],
                    ["SIMPLICITY", "GAS_EXTRACTION", "IN", 1, 2014, 1],
                    ["SIMPLICITY", "GAS_EXTRACTION", "SD", 1, 2014, 1],
                    ["SIMPLICITY", "GAS_EXTRACTION", "SN", 1, 2014, 1],
                    ["SIMPLICITY", "GAS_EXTRACTION", "WD", 1, 2014, 1],
                    ["SIMPLICITY", "GAS_EXTRACTION", "WN", 1, 2014, 1],
                    ["SIMPLICITY", "DUMMY", "ID", 1, 2014, 1],
                    ["SIMPLICITY", "DUMMY", "IN", 1, 2014, 1],
                    ["SIMPLICITY", "DUMMY", "SD", 1, 2014, 1],
                    ["SIMPLICITY", "DUMMY", "SN", 1, 2014, 1],
                    ["SIMPLICITY", "DUMMY", "WD", 1, 2014, 1],
                    ["SIMPLICITY", "DUMMY", "WN", 1, 2014, 1],
                ],
                columns=[
                    "REGION",
                    "TECHNOLOGY",
                    "TIMESLICE",
                    "MODE_OF_OPERATION",
                    "YEAR",
                    "VALUE",
                ],
            ).set_index(
                ["REGION", "TECHNOLOGY", "TIMESLICE", "MODE_OF_OPERATION", "YEAR"]
            )
        }
        input_data = {
            "EmissionActivityRatio": pd.DataFrame(
                data=[["SIMPLICITY", "GAS_EXTRACTION", "CO2", 1, 2014, 1.0]],
                columns=[
                    "REGION",
                    "TECHNOLOGY",
                    "EMISSION",
                    "MODE_OF_OPERATION",
                    "YEAR",
                    "VALUE",
                ],
            ).set_index(
                ["REGION", "TECHNOLOGY", "EMISSION", "MODE_OF_OPERATION", "YEAR"]
            ),
            "YearSplit": pd.DataFrame(
                data=[
                    ["ID", 2014, 0.1667],
                    ["IN", 2014, 0.0833],
                    ["SD", 2014, 0.1667],
                    ["SN", 2014, 0.0833],
                    ["WD", 2014, 0.3333],
                    ["WN", 2014, 0.1667],
                ],
                columns=["TIMESLICE", "YEAR", "VALUE"],
            ).set_index(["TIMESLICE", "YEAR"]),
        }

        expected = pd.DataFrame(
            data=[["SIMPLICITY", "CO2", 2014, 1.0]],
            columns=["REGION", "EMISSION", "YEAR", "VALUE"],
        ).set_index(["REGION", "EMISSION", "YEAR"])

        cbc_reader = ReadCbc(user_config)
        actual = cbc_reader.calculate_results(cbc_results, input_data)
        assert isinstance(actual, dict)
        pd.testing.assert_frame_equal(actual["AnnualEmissions"], expected)

    def test_solution_to_dataframe(self, user_config):
        input_file = self.total_cost_cbc
        reader = ReadCbc(user_config)
        with StringIO(input_file) as file_buffer:
            actual = reader.read(file_buffer)
        expected = self.total_cost_otoole_df
        pd.testing.assert_frame_equal(
            actual[0]["TotalDiscountedCost"], expected["TotalDiscountedCost"]
        )

    cbc_infeasible = dedent(
        """header
381191 RateOfActivity(GLOBAL,S4D24,INRNGIM00,1,2041)                           0             0
381191 RateOfActivity(GLOBAL,S4D24,INRNGIM00,1,2042)                           -7.7011981e-07             0.024001857
381192 RateOfActivity(GLOBAL,S1D1,INRNGIM00,1,2043)                            -3.6128354e-06             0.022858911
381199 RateOfActivity(GLOBAL,S1D8,INRNGIM00,1,2043)                            -3.1111316e-06             0.022858911
381200 RateOfActivity(GLOBAL,S1D9,INRNGIM00,1,2043)                            -8.2325306e-07             0.022858911
381201 RateOfActivity(GLOBAL,S1D10,INRNGIM00,1,2043)                           -3.1112991e-06             0.022858911
**  381218 RateOfActivity(GLOBAL,S2D3,INRNGIM00,1,2043)                            -1.6357402e-06             0.022858911
**  381221 RateOfActivity(GLOBAL,S2D6,INRNGIM00,1,2043)                            -3.1111969e-06             0.022858911
**  381229 RateOfActivity(GLOBAL,S2D14,INRNGIM00,1,2043)                           -1.3925924e-07             0.010964295
"""
    )

    def test_manage_infeasible_variables(self, user_config):
        input_file = self.cbc_infeasible
        reader = ReadCbc(user_config)
        with StringIO(input_file) as file_buffer:
            actual = reader._convert_to_dataframe(file_buffer)
        expected = pd.DataFrame(
            [
                ["RateOfActivity", "GLOBAL,S4D24,INRNGIM00,1,2041", 0],
                ["RateOfActivity", "GLOBAL,S4D24,INRNGIM00,1,2042", -7.7011981e-07],
                ["RateOfActivity", "GLOBAL,S1D1,INRNGIM00,1,2043", -3.6128354e-06],
                ["RateOfActivity", "GLOBAL,S1D8,INRNGIM00,1,2043", -3.1111316e-06],
                ["RateOfActivity", "GLOBAL,S1D9,INRNGIM00,1,2043", -8.2325306e-07],
                ["RateOfActivity", "GLOBAL,S1D10,INRNGIM00,1,2043", -3.1112991e-06],
                ["RateOfActivity", "GLOBAL,S2D3,INRNGIM00,1,2043", -1.6357402e-06],
                ["RateOfActivity", "GLOBAL,S2D6,INRNGIM00,1,2043", -3.1111969e-06],
                ["RateOfActivity", "GLOBAL,S2D14,INRNGIM00,1,2043", -1.3925924e-07],
            ],
            columns=["Variable", "Index", "Value"],
        )
        pd.testing.assert_frame_equal(actual, expected)


class TestCleanOnRead:
    """Tests that a data is cleaned and indexed upon reading"""

    def test_index_dtypes_available(self, user_config):
        reader = ReadMemory({}, user_config=user_config)
        config = reader.user_config
        assert "index_dtypes" in config["AccumulatedAnnualDemand"].keys()
        actual = config["AccumulatedAnnualDemand"]["index_dtypes"]
        assert actual == {
            "REGION": "str",
            "FUEL": "str",
            "YEAR": "int",
            "VALUE": "float",
        }

    def test_remove_empty_lines(self, user_config):

        data = [
            ["SIMPLICITY", "ETH", 2014, 1.0],
            [],
            ["SIMPLICITY", "ETH", 2015, 1.03],
            [],
        ]
        df = pd.DataFrame(data=data, columns=["REGION", "FUEL", "YEAR", "VALUE"])
        parameters = {"AccumulatedAnnualDemand": df}

        reader = ReadMemory(parameters, user_config=user_config)
        actual, _ = reader.read()

        expected = {
            "AccumulatedAnnualDemand": pd.DataFrame(
                data=[
                    ["SIMPLICITY", "ETH", 2014, 1.0],
                    ["SIMPLICITY", "ETH", 2015, 1.03],
                ],
                columns=["REGION", "FUEL", "YEAR", "VALUE"],
            )
            .astype({"REGION": str, "FUEL": str, "YEAR": int, "VALUE": float})
            .set_index(["REGION", "FUEL", "YEAR"])
        }

        assert "AccumulatedAnnualDemand" in actual.keys()
        pd.testing.assert_frame_equal(
            actual["AccumulatedAnnualDemand"], expected["AccumulatedAnnualDemand"]
        )

    def test_change_types(self, user_config):

        data = [
            ["SIMPLICITY", "ETH", 2014.0, 1],
            ["SIMPLICITY", "ETH", 2015.0, 1],
        ]
        df = pd.DataFrame(
            data=data, columns=["REGION", "FUEL", "YEAR", "VALUE"]
        ).astype({"REGION": str, "FUEL": str, "YEAR": float, "VALUE": int})
        parameters = {"AccumulatedAnnualDemand": df}

        reader = ReadMemory(parameters, user_config=user_config)
        actual, _ = reader.read()

        expected = {
            "AccumulatedAnnualDemand": pd.DataFrame(
                data=[
                    ["SIMPLICITY", "ETH", 2014, 1.0],
                    ["SIMPLICITY", "ETH", 2015, 1.0],
                ],
                columns=["REGION", "FUEL", "YEAR", "VALUE"],
            )
            .astype({"REGION": str, "FUEL": str, "YEAR": int, "VALUE": float})
            .set_index(["REGION", "FUEL", "YEAR"])
        }

        assert "AccumulatedAnnualDemand" in actual.keys()
        pd.testing.assert_frame_equal(
            actual["AccumulatedAnnualDemand"], expected["AccumulatedAnnualDemand"]
        )


class TestReadMemoryStrategy:
    def test_read_memory(self, user_config):

        data = [
            ["SIMPLICITY", "ETH", 2014, 1.0],
            ["SIMPLICITY", "RAWSUG", 2014, 0.5],
            ["SIMPLICITY", "ETH", 2015, 1.03],
            ["SIMPLICITY", "RAWSUG", 2015, 0.51],
        ]
        df = pd.DataFrame(data=data, columns=["REGION", "FUEL", "YEAR", "VALUE"])
        parameters = {"AccumulatedAnnualDemand": df}

        actual, default_values = ReadMemory(parameters, user_config).read()

        expected = {
            "AccumulatedAnnualDemand": pd.DataFrame(
                data=data, columns=["REGION", "FUEL", "YEAR", "VALUE"]
            ).set_index(["REGION", "FUEL", "YEAR"])
        }

        assert "AccumulatedAnnualDemand" in actual.keys()
        pd.testing.assert_frame_equal(
            actual["AccumulatedAnnualDemand"], expected["AccumulatedAnnualDemand"]
        )

    def test_read_memory_user_config(self, user_config):

        data = [
            ["SIMPLICITY", "ETH", 2014, 1.0],
            ["SIMPLICITY", "RAWSUG", 2014, 0.5],
            ["SIMPLICITY", "ETH", 2015, 1.03],
            ["SIMPLICITY", "RAWSUG", 2015, 0.51],
        ]
        df = pd.DataFrame(data=data, columns=["REGION", "FUEL", "YEAR", "VALUE"])
        parameters = {"AccumulatedAnnualDemand": df}

        actual, default_values = ReadMemory(parameters, user_config=user_config).read()

        expected = {
            "AccumulatedAnnualDemand": pd.DataFrame(
                data=data, columns=["REGION", "FUEL", "YEAR", "VALUE"]
            ).set_index(["REGION", "FUEL", "YEAR"])
        }

        assert "AccumulatedAnnualDemand" in actual.keys()
        pd.testing.assert_frame_equal(
            actual["AccumulatedAnnualDemand"], expected["AccumulatedAnnualDemand"]
        )

        assert default_values["AccumulatedAnnualDemand"] == 0


class TestConfig:
    def test_read_config(self, user_config):

        read = ReadDatafile(user_config=user_config)

        actual = read.user_config
        expected = {
            "default": 0,
            "dtype": "float",
            "indices": ["REGION", "FUEL", "YEAR"],
            "type": "param",
            "index_dtypes": {
                "FUEL": "str",
                "REGION": "str",
                "VALUE": "float",
                "YEAR": "int",
            },
        }
        assert actual["AccumulatedAnnualDemand"] == expected


class TestReadDatafile:
    def test_amply(self):

        Amply(
            """set REGION;
            # set REGION := SIMPLICITY;
            set TECHNOLOGY;
            set TECHNOLOGY := ETHPLANT GAS_EXTRACTION;
            set MODE_OF_OPERATION;
            set MODE_OF_OPERATION := 1 2;
            set YEAR;
            set YEAR := 2014;
            end;"""
        )

    def test_convert_amply_to_dataframe(self, user_config):

        config = {
            "VariableCost": {
                "type": "param",
                "indices": ["REGION", "TECHNOLOGY", "MODE_OF_OPERATION", "YEAR"],
                "dtype": "float",
                "default": 0,
            },
            "REGION": {"type": "set", "dtype": "str"},
            "YEAR": {"dtype": "int", "type": "set"},
            "MODE_OF_OPERATION": {"dtype": "int", "type": "set"},
            "TECHNOLOGY": {"dtype": "str", "type": "set"},
        }

        amply = Amply(
            """set REGION;
                        set REGION := SIMPLICITY;
                        set TECHNOLOGY;
                        set TECHNOLOGY := ETHPLANT GAS_EXTRACTION;
                        set MODE_OF_OPERATION;
                        set MODE_OF_OPERATION := 1 2;
                        set YEAR;
                        set YEAR := 2014;"""
        )
        amply.load_string(
            "param VariableCost {REGION,TECHNOLOGY,MODE_OF_OPERATION,YEAR};"
        )
        #     amply.load_string("""param default 0 : VariableCost :=
        # SIMPLICITY ETHPLANT 1 2014 2.89
        # SIMPLICITY ETHPLANT 2 2014 999999.0
        # SIMPLICITY GAS_EXTRACTION 1 2014 7.5
        # SIMPLICITY GAS_EXTRACTION 2 2014 999999.0""")
        amply.load_string(
            """
    param VariableCost default 0.0001 :=
    [SIMPLICITY,ETHPLANT,*,*]:
    2014 :=
    1 2.89
    2 999999.0
    [SIMPLICITY,GAS_EXTRACTION,*,*]:
    2014 :=
    1 7.5
    2 999999.0;"""
        )

        read = ReadDatafile(user_config=user_config)

        actual = read._convert_amply_to_dataframe(amply, config)
        expected = pd.DataFrame(
            data=[
                ["SIMPLICITY", "ETHPLANT", 1, 2014, 2.89],
                ["SIMPLICITY", "ETHPLANT", 2, 2014, 999999.0],
                ["SIMPLICITY", "GAS_EXTRACTION", 1, 2014, 7.5],
                ["SIMPLICITY", "GAS_EXTRACTION", 2, 2014, 999999.0],
            ],
            columns=["REGION", "TECHNOLOGY", "MODE_OF_OPERATION", "YEAR", "VALUE"],
        )
        pd.testing.assert_frame_equal(actual["VariableCost"], expected)

    def test_convert_amply_data_to_list_of_lists(self, user_config):

        data = {
            "SIMPLICITY": {
                "ETHPLANT": {1.0: {2014.0: 2.89}, 2.0: {2014.0: 999999.0}},
                "GAS_EXTRACTION": {1.0: {2014.0: 7.5}, 2.0: {2014.0: 999999.0}},
            }
        }
        expected = [
            ["SIMPLICITY", "ETHPLANT", 1.0, 2014.0, 2.89],
            ["SIMPLICITY", "ETHPLANT", 2.0, 2014.0, 999999.0],
            ["SIMPLICITY", "GAS_EXTRACTION", 1.0, 2014.0, 7.5],
            ["SIMPLICITY", "GAS_EXTRACTION", 2.0, 2014.0, 999999.0],
        ]
        read = ReadDatafile(user_config=user_config)
        actual = read._convert_amply_data_to_list(data)
        assert actual == expected

    def test_load_parameters(self, user_config):

        config = {"TestParameter": {"type": "param", "indices": ["index1", "index2"]}}
        read = ReadDatafile(user_config=user_config)
        actual = read._load_parameter_definitions(config)
        expected = "param TestParameter {index1,index2};\n"
        assert actual == expected

    def test_load_sets(self, user_config):

        config = {"TestSet": {"type": "set"}}

        read = ReadDatafile(user_config=user_config)
        actual = read._load_parameter_definitions(config)
        expected = "set TestSet;\n"
        assert actual == expected

    def test_catch_error_no_parameter(self, caplog, user_config):
        """Fix for https://github.com/OSeMOSYS/otoole/issues/70 where parameter in
        datafile but not in config causes error.  Instead, throw warning (and advise
        that user should use a custom configuration).
        """
        read = ReadDatafile(user_config=user_config)
        config = read.user_config
        amply_datafile = amply = Amply(
            """set REGION;
            set TECHNOLOGY;
            set MODE_OF_OPERATION;
            set YEAR;"""
        )
        amply.load_string("""param ResultsPath := 'test_path';""")
        read._convert_amply_to_dataframe(amply_datafile, config)
        assert (
            "Parameter ResultsPath could not be found in the configuration."
            in caplog.text
        )


class TestReadExcel:
    def test_read_excel_yearsplit(self, user_config):
        """ """
        spreadsheet = os.path.join("tests", "fixtures", "combined_inputs.xlsx")
        reader = ReadExcel(user_config=user_config)
        actual, _ = reader.read(spreadsheet)
        data = [
            ["IW0016", 2017, 0.238356164],
            ["IW0016", 2018, 0.238356164],
            ["IW0016", 2019, 0.238356164],
            ["IW1624", 2017, 0.119178082],
            ["IW1624", 2018, 0.119178082],
            ["IW1624", 2019, 0.119178082],
            ["IH0012", 2017, 0.071232876],
            ["IH0012", 2018, 0.071232876],
            ["IH0012", 2019, 0.071232876],
        ]
        expected = pd.DataFrame(data, columns=["TIMESLICE", "YEAR", "VALUE"]).set_index(
            ["TIMESLICE", "YEAR"]
        )

        assert "YearSplit" in actual

        index = [
            ("IW0016", 2017),
            ("IW0016", 2018),
            ("IW0016", 2019),
            ("IW1624", 2017),
            ("IW1624", 2018),
            ("IW1624", 2019),
            ("IH0012", 2017),
            ("IH0012", 2018),
            ("IH0012", 2019),
        ]

        assert actual["YearSplit"].index.names == ["TIMESLICE", "YEAR"]
        actual_data = actual["YearSplit"].loc[index, "VALUE"]

        expected = [
            0.238356164,
            0.238356164,
            0.238356164,
            0.119178082,
            0.119178082,
            0.119178082,
            0.071232876,
            0.071232876,
            0.071232876,
        ]

        assert (actual_data == expected).all()

    def test_read_excel_discount_rate(self, user_config):
        """Tests that parameters not in excel are saved in datastore"""

        spreadsheet = os.path.join("tests", "fixtures", "combined_inputs.xlsx")
        xl = pd.ExcelFile(spreadsheet, engine="openpyxl")

        # checks that fixture does not contian discount rate data
        assert "DiscountRateIdv" not in xl.sheet_names

        reader = ReadExcel(user_config=user_config)
        actual, _ = reader.read(spreadsheet)

        # checks that discount rate has data after reading in excel data
        assert "DiscountRateIdv" in actual
        assert actual["DiscountRateIdv"].empty

    def test_narrow_parameters(self, user_config):
        data = [
            ["IW0016", 0.238356164, 0.238356164, 0.238356164],
            ["IW1624", 0.119178082, 0.119178082, 0.119178082],
            ["IH0012", 0.071232876, 0.071232876, 0.071232876],
        ]
        df = pd.DataFrame(data, columns=["TIMESLICE", 2017, 2018, 2019])
        config_details = ["TIMESLICE", "YEAR"]
        name = "YearSplit"

        reader = ReadExcel(user_config=user_config)
        actual = reader._check_parameter(df, config_details, name)
        data = [
            ["IW0016", 2017, 0.238356164],
            ["IW1624", 2017, 0.119178082],
            ["IH0012", 2017, 0.071232876],
            ["IW0016", 2018, 0.238356164],
            ["IW1624", 2018, 0.119178082],
            ["IH0012", 2018, 0.071232876],
            ["IW0016", 2019, 0.238356164],
            ["IW1624", 2019, 0.119178082],
            ["IH0012", 2019, 0.071232876],
        ]
        expected = (
            pd.DataFrame(data, columns=["TIMESLICE", "YEAR", "VALUE"])
            .astype({"YEAR": "object"})
            .set_index(["TIMESLICE", "YEAR"])
        )
        pd.testing.assert_frame_equal(actual, expected)

    def test_check_index(self, user_config):

        data = [
            ["IW0016", 2017, 0.238356164],
            ["IW0016", 2018, 0.238356164],
            ["IW0016", 2019, 0.238356164],
            ["IW1624", 2017, 0.119178082],
            ["IW1624", 2018, 0.119178082],
            ["IW1624", 2019, 0.119178082],
            ["IH0012", 2017, 0.071232876],
            ["IH0012", 2018, 0.071232876],
            ["IH0012", 2019, 0.071232876],
        ]
        df = {
            "YearSplit": pd.DataFrame(data, columns=["TIMESLICE", "YEAR", "VALUE"])
            .astype({"YEAR": object})
            .set_index(["TIMESLICE", "YEAR"])
        }
        reader = ReadExcel(user_config=user_config)
        actual = reader._check_index(df)
        expected = {
            "YearSplit": pd.DataFrame(
                data, columns=["TIMESLICE", "YEAR", "VALUE"]
            ).set_index(["TIMESLICE", "YEAR"])
        }
        pd.testing.assert_frame_equal(actual["YearSplit"], expected["YearSplit"])


class TestReadCSV:
    accumulated_annual_demand_df = pd.DataFrame(
        [
            ["SIMPLICITY", "ETH", 2014, 1.0],
            ["SIMPLICITY", "RAWSUG", 2014, 0.5],
            ["SIMPLICITY", "ETH", 2015, 1.03],
            ["SIMPLICITY", "RAWSUG", 2015, 0.51],
            ["SIMPLICITY", "ETH", 2016, 1.061],
            ["SIMPLICITY", "RAWSUG", 2016, 0.519],
        ],
        columns=["REGION", "FUEL", "YEAR", "VALUE"],
    )
    availability_factor_df = pd.DataFrame(
        columns=["REGION", "TECHNOLOGY", "YEAR", "VALUE"]
    )

    test_data = [
        ("AccumulatedAnnualDemand", accumulated_annual_demand_df),
        ("AvailabilityFactor", availability_factor_df),
    ]

    @mark.parametrize("parameter, expected", test_data, ids=["full", "empty"])
    def test_get_input_data_empty(self, user_config, parameter, expected):
        reader = ReadCsv(user_config=user_config)
        filepath = os.path.join("tests", "fixtures", "data")
        details = user_config[parameter]
        actual = reader._get_input_data(filepath, parameter, details)

        pd.testing.assert_frame_equal(actual, expected)

    def test_read_default_values_csv_fails(self, user_config, tmp_path):
        f = tmp_path / "input/default_values.csv"
        f.parent.mkdir()
        f.touch()

        reader = ReadCsv(user_config=user_config)
        with raises(OtooleDeprecationError):
            reader._check_for_default_values_csv(f)

    def test_read_default_values_csv(self, user_config):
        filepath = os.path.join(
            "tests", "fixtures", "data", "AccumulatedAnnualDemand.csv"
        )
        reader = ReadCsv(user_config=user_config)
        actual = reader._check_for_default_values_csv(filepath)
        expected = None
        assert actual == expected


class TestReadTabular:
    """Methods shared for csv and excel"""

    test_data = [
        (True, ["REGION", "TECHNOLOGY"], {}),
        (
            False,
            ["REGION", "TECHNOLOGY"],
            {"REGION": str.strip, "TECHNOLOGY": str.strip},
        ),
    ]

    @mark.parametrize(
        "keep_whitespace, indices, expected",
        test_data,
        ids=["create_empty", "create_full"],
    )
    def test_whitespace_converter(
        self, user_config, keep_whitespace, indices, expected
    ):
        reader = ReadCsv(user_config=user_config, keep_whitespace=keep_whitespace)
        actual = reader._whitespace_converter(indices)
        assert actual == expected



"""Visualise different aspects of an OSeMOSYS model and data

Provides the following commands::

    otoole viz res <input_data_format> <path_to_input_data> <path_to_res_image> <path_to_user_config>

``otoole viz res`` generates an image of the reference energy system

"""
from .res import create_res

__all__ = ["create_res"]



from typing import Any, Dict, TextIO, Tuple, Union

import pandas as pd
from pandas.testing import assert_frame_equal
from pytest import fixture, mark, raises

from otoole.input import ReadStrategy, WriteStrategy


@fixture
def year():
    return pd.DataFrame(data=[2014, 2015, 2016], columns=["VALUE"])


@fixture
def region():
    return pd.DataFrame(data=["SIMPLICITY"], columns=["VALUE"])


@fixture
def technology():
    return pd.DataFrame(data=["NGCC", "HYD1"], columns=["VALUE"])


@fixture()
def simple_default_values():
    default_values = {}
    default_values["CapitalCost"] = -1.0
    default_values["DiscountRate"] = 0.25
    default_values["NewCapacity"] = 20
    return default_values


@fixture
def simple_input_data(region, year, technology):
    return {
        "REGION": region,
        "TECHNOLOGY": technology,
        "YEAR": year,
    }


# To instantiate abstract class WriteStrategy
class DummyWriteStrategy(WriteStrategy):
    def _header(self) -> Union[TextIO, Any]:
        raise NotImplementedError()

    def _write_parameter(
        self,
        df: pd.DataFrame,
        parameter_name: str,
        handle: TextIO,
        default: float,
        **kwargs
    ) -> pd.DataFrame:
        raise NotImplementedError()

    def _write_set(self, df: pd.DataFrame, set_name, handle: TextIO) -> pd.DataFrame:
        raise NotImplementedError()

    def _footer(self, handle: TextIO):
        raise NotImplementedError()


# To instantiate abstract class ReadStrategy
class DummyReadStrategy(ReadStrategy):
    def read(
        self, filepath: Union[str, TextIO], **kwargs
    ) -> Tuple[Dict[str, pd.DataFrame], Dict[str, Any]]:
        raise NotImplementedError()


class TestExpandDefaults:

    year = pd.DataFrame(data=[2014, 2015, 2016], columns=["VALUE"])
    region = pd.DataFrame(data=["SIMPLICITY"], columns=["VALUE"])
    technology = pd.DataFrame(data=["NGCC", "HYD1"], columns=["VALUE"])

    def input_data_multi_index_no_defaults(region, technology, year):
        capex_in = pd.DataFrame(
            [
                ["SIMPLICITY", "HYD1", 2014, 2000],
                ["SIMPLICITY", "HYD1", 2015, 1500],
                ["SIMPLICITY", "HYD1", 2016, 1000],
                ["SIMPLICITY", "NGCC", 2014, 1000],
                ["SIMPLICITY", "NGCC", 2015, 900],
                ["SIMPLICITY", "NGCC", 2016, 800],
            ],
            columns=["REGION", "TECHNOLOGY", "YEAR", "VALUE"],
        ).set_index(["REGION", "TECHNOLOGY", "YEAR"])
        capex_out = capex_in.copy()
        capex_out["VALUE"] = capex_out["VALUE"].astype(float)

        data = {
            "CapitalCost": capex_in,
            "TECHNOLOGY": technology,
            "YEAR": year,
            "REGION": region,
        }
        return data, "CapitalCost", capex_out

    def input_data_multi_index(region, technology, year):
        capex_in = pd.DataFrame(
            [
                ["SIMPLICITY", "NGCC", 2014, 1000],
                ["SIMPLICITY", "NGCC", 2015, 900],
                ["SIMPLICITY", "HYD1", 2015, 1500],
                ["SIMPLICITY", "HYD1", 2016, 1000],
            ],
            columns=["REGION", "TECHNOLOGY", "YEAR", "VALUE"],
        ).set_index(["REGION", "TECHNOLOGY", "YEAR"])
        capex_out = pd.DataFrame(
            [
                ["SIMPLICITY", "HYD1", 2014, -1],
                ["SIMPLICITY", "HYD1", 2015, 1500],
                ["SIMPLICITY", "HYD1", 2016, 1000],
                ["SIMPLICITY", "NGCC", 2014, 1000],
                ["SIMPLICITY", "NGCC", 2015, 900],
                ["SIMPLICITY", "NGCC", 2016, -1],
            ],
            columns=["REGION", "TECHNOLOGY", "YEAR", "VALUE"],
        ).set_index(["REGION", "TECHNOLOGY", "YEAR"])
        capex_out["VALUE"] = capex_out["VALUE"].astype(float)

        data = {
            "CapitalCost": capex_in,
            "TECHNOLOGY": technology,
            "YEAR": year,
            "REGION": region,
        }
        return data, "CapitalCost", capex_out

    def input_data_multi_index_empty(region, technology, year):
        capex_in = pd.DataFrame(
            [],
            columns=["REGION", "TECHNOLOGY", "YEAR", "VALUE"],
        ).set_index(["REGION", "TECHNOLOGY", "YEAR"])
        capex_out = pd.DataFrame(
            [
                ["SIMPLICITY", "HYD1", 2014, -1],
                ["SIMPLICITY", "HYD1", 2015, -1],
                ["SIMPLICITY", "HYD1", 2016, -1],
                ["SIMPLICITY", "NGCC", 2014, -1],
                ["SIMPLICITY", "NGCC", 2015, -1],
                ["SIMPLICITY", "NGCC", 2016, -1],
            ],
            columns=["REGION", "TECHNOLOGY", "YEAR", "VALUE"],
        ).set_index(["REGION", "TECHNOLOGY", "YEAR"])
        capex_out["VALUE"] = capex_out["VALUE"].astype(float)

        data = {
            "CapitalCost": capex_in,
            "TECHNOLOGY": technology,
            "YEAR": year,
            "REGION": region,
        }
        return data, "CapitalCost", capex_out

    def input_data_single_index(region):
        discount_rate_in = pd.DataFrame(
            [["SIMPLICITY", 0.05]], columns=["REGION", "VALUE"]
        ).set_index(["REGION"])
        discount_rate_out = discount_rate_in.copy()
        discount_rate_out["VALUE"] = discount_rate_out["VALUE"].astype(float)

        data = {
            "DiscountRate": discount_rate_in,
            "REGION": region,
        }
        return data, "DiscountRate", discount_rate_out

    def input_data_single_index_empty(region):
        discount_rate_in = pd.DataFrame([], columns=["REGION", "VALUE"]).set_index(
            ["REGION"]
        )
        discount_rate_out = pd.DataFrame(
            [["SIMPLICITY", 0.25]], columns=["REGION", "VALUE"]
        ).set_index(["REGION"])
        discount_rate_out["VALUE"] = discount_rate_out["VALUE"].astype(float)

        data = {
            "DiscountRate": discount_rate_in,
            "TECHNOLOGY": technology,
            "YEAR": year,
            "REGION": region,
        }
        return data, "DiscountRate", discount_rate_out

    @fixture
    def result_data(region):
        new_capacity_in = pd.DataFrame(
            [
                ["SIMPLICITY", "HYD1", 2015, 100],
                ["SIMPLICITY", "HYD1", 2016, 0.1],
                ["SIMPLICITY", "NGCC", 2014, 0.5],
                ["SIMPLICITY", "NGCC", 2015, 100],
            ],
            columns=["REGION", "TECHNOLOGY", "YEAR", "VALUE"],
        ).set_index(["REGION", "TECHNOLOGY", "YEAR"])
        new_capacity_out = pd.DataFrame(
            [
                ["SIMPLICITY", "HYD1", 2014, 20],
                ["SIMPLICITY", "HYD1", 2015, 100],
                ["SIMPLICITY", "HYD1", 2016, 0.1],
                ["SIMPLICITY", "NGCC", 2014, 0.5],
                ["SIMPLICITY", "NGCC", 2015, 100],
                ["SIMPLICITY", "NGCC", 2016, 20],
            ],
            columns=["REGION", "TECHNOLOGY", "YEAR", "VALUE"],
        ).set_index(["REGION", "TECHNOLOGY", "YEAR"])

        data = {
            "NewCapacity": new_capacity_in,
        }
        return data, "NewCapacity", new_capacity_out

    parameter_test_data = [
        input_data_multi_index_no_defaults(region, technology, year),
        input_data_multi_index(region, technology, year),
        input_data_multi_index_empty(region, technology, year),
        input_data_single_index(region),
        input_data_single_index_empty(region),
    ]
    parameter_test_data_ids = [
        "multi_index_no_defaluts",
        "multi_index",
        "multi_index_empty",
        "single_index",
        "single_index_empty",
    ]

    @mark.parametrize(
        "input_data,parameter,expected",
        parameter_test_data,
        ids=parameter_test_data_ids,
    )
    def test_expand_parmaters_defaults(
        self, user_config, simple_default_values, input_data, parameter, expected
    ):
        write_strategy = DummyWriteStrategy(
            user_config=user_config, default_values=simple_default_values
        )
        actual = write_strategy._expand_defaults(
            input_data, write_strategy.default_values
        )
        assert_frame_equal(actual[parameter], expected)

    def test_expand_result_defaults(
        self, user_config, simple_default_values, simple_input_data, result_data
    ):
        write_strategy = DummyWriteStrategy(
            user_config=user_config, default_values=simple_default_values
        )
        actual = write_strategy._expand_defaults(
            result_data[0], write_strategy.default_values, input_data=simple_input_data
        )
        assert_frame_equal(actual[result_data[1]], result_data[2])

    def test_expand_defaults_exception(
        self, user_config, simple_default_values, result_data
    ):
        write_strategy = DummyWriteStrategy(
            user_config=user_config, default_values=simple_default_values
        )
        with raises(KeyError):
            write_strategy._expand_defaults(
                result_data[0], write_strategy.default_values
            )


class TestReadStrategy:
    missing_input_test_data = (
        (
            "param",
            "AccumulatedAnnualDemand",
            pd.DataFrame(columns=["REGION", "FUEL", "YEAR", "VALUE"]).set_index(
                ["REGION", "FUEL", "YEAR"]
            ),
        ),
        ("set", "REGION", pd.DataFrame(columns=["VALUE"])),
    )

    @mark.parametrize(
        "config_type, test_value, expected",
        missing_input_test_data,
        ids=["param", "set"],
    )
    def test_get_missing_input_dataframes(
        self, user_config, config_type, test_value, expected
    ):
        input_data = {}
        reader = DummyReadStrategy(user_config)
        input_data = reader._get_missing_input_dataframes(input_data, config_type)
        actual = input_data[test_value]
        pd.testing.assert_frame_equal(actual, expected)

    def test_get_missing_input_dataframes_excpetion(self, user_config):
        input_data = {}
        reader = DummyReadStrategy(user_config)
        with raises(ValueError):
            reader._get_missing_input_dataframes(input_data, config_type="not_valid")
